"""
Multi-part upload as a graph
"""

from __future__ import annotations

from functools import partial
from typing import (
    TYPE_CHECKING,
    Any,
    Iterable,
    Iterator,
    Optional,
    Protocol,
    Union,
)

if TYPE_CHECKING:
    import dask.bag
    from dask.delayed import Delayed

__all__ = [
    "SomeData",
    "PartsWriter",
    "MPUChunk",
    "mpu_write",
]

SomeData = Union[bytes, bytearray]


class PartsWriter(Protocol):
    """Protocol for labeled parts data writer."""

    def __call__(self, part: int, data: SomeData) -> dict[str, Any]: ...

    def finalise(self, parts: list[dict[str, Any]]) -> Any: ...

    @property
    def min_write_sz(self) -> int: ...

    @property
    def max_write_sz(self) -> int: ...

    @property
    def min_part(self) -> int: ...

    @property
    def max_part(self) -> int: ...


class MPUChunk:
    """
    chunk cache and writer
    """

    # pylint: disable=too-many-arguments,too-many-instance-attributes

    __slots__ = (
        "nextPartId",
        "write_credits",
        "data",
        "left_data",
        "parts",
        "observed",
        "is_final",
        "lhs_keep",
    )

    def __init__(
        self,
        partId: int,
        write_credits: int,
        data: Optional[bytearray] = None,
        left_data: Optional[bytearray] = None,
        parts: Optional[list[dict[str, Any]]] = None,
        observed: Optional[list[tuple[int, Any]]] = None,
        is_final: bool = False,
        lhs_keep: int = 0,
    ) -> None:
        self.nextPartId = partId
        self.write_credits = write_credits
        self.data = bytearray() if data is None else data
        self.left_data = bytearray() if left_data is None else left_data
        self.parts: list[dict[str, Any]] = [] if parts is None else parts
        self.observed: list[tuple[int, Any]] = [] if observed is None else observed
        self.is_final = is_final
        self.lhs_keep = lhs_keep
        # if supplying data must also supply observed
        assert data is None or (observed is not None and len(observed) > 0)

    def __dask_tokenize__(self):
        return (
            "MPUChunk",
            self.nextPartId,
            self.write_credits,
            self.data,
            self.left_data,
            self.parts,
            self.observed,
            self.is_final,
        )

    def __repr__(self) -> str:
        s = f"MPUChunk: {self.nextPartId}#{self.write_credits} cache: {len(self.data)}"
        if self.observed:
            s = f"{s} observed[{len(self.observed)}]"
        if self.parts:
            s = f"{s} parts: [{len(self.parts)}]"
        if self.is_final:
            s = f"{s} final"
        return s

    def append(self, data: SomeData, chunk_id: Any = None):
        sz = len(data)
        self.observed.append((sz, chunk_id))
        self.data += data

    @property
    def started_write(self) -> bool:
        return len(self.parts) > 0

    @staticmethod
    def merge(
        lhs: "MPUChunk",
        rhs: "MPUChunk",
        write: Optional[PartsWriter] = None,
    ) -> "MPUChunk":
        """
        If ``write=`` is not provided but flush is needed, RuntimeError will be raised.
        """
        if not rhs.started_write:
            # no writes on the right
            # Just append
            assert len(rhs.left_data) == 0
            assert len(rhs.parts) == 0

            return MPUChunk(
                lhs.nextPartId,
                lhs.write_credits + rhs.write_credits,
                lhs.data + rhs.data,
                lhs.left_data,
                lhs.parts,
                lhs.observed + rhs.observed,
                rhs.is_final,
                lhs.lhs_keep,
            )

        # Flush `lhs.data + rhs.left_data` if we can
        #  or else move it into .left_data
        lhs.flush_rhs(write, rhs.left_data)

        return MPUChunk(
            rhs.nextPartId,
            rhs.write_credits,
            rhs.data,
            lhs.left_data,
            lhs.parts + rhs.parts,
            lhs.observed + rhs.observed,
            rhs.is_final,
            lhs.lhs_keep,
        )

    def flush_rhs(
        self, write: Optional[PartsWriter], extra_data: Optional[bytearray] = None
    ) -> int:
        data = self.data
        if extra_data is not None and len(extra_data):
            data += extra_data

        def _flush_data(pw: PartsWriter):
            assert pw.min_part <= self.nextPartId <= pw.max_part

            _data = data
            if not self.started_write and self.lhs_keep > 0:
                self.left_data = bytearray(_data[: self.lhs_keep])
                _data = data[self.lhs_keep :]

            part = pw(self.nextPartId, _data)

            self.parts.append(part)
            self.data = bytearray()
            self.nextPartId += 1
            self.write_credits -= 1
            return len(_data)

        def can_flush(pw: PartsWriter):
            if self.write_credits < 1:
                return False
            if self.started_write:
                return self.is_final or len(data) >= pw.min_write_sz
            if self.is_final:
                return len(data) > self.lhs_keep
            return len(data) - self.lhs_keep >= pw.min_write_sz

        if self.started_write:
            # When starting to write we ensure that there is always enough
            # data and write credits left to flush the remainder
            #
            # User must have provided `write` function
            if write is None:
                raise RuntimeError("Flush required but no writer provided")

            assert can_flush(write)
            return _flush_data(write)

        # Haven't started writing yet
        # - Flush if possible and writer is provided
        # - OR just move all the data to .left_data section
        if write is not None and can_flush(write):
            return _flush_data(write)

        self.left_data, self.data = self.left_data + data, bytearray()
        return 0

    def flush(
        self,
        write: PartsWriter,
        leftPartId: Optional[int] = None,
        finalise: bool = True,
    ) -> tuple[int, Any]:
        rr = None
        if not self.started_write:
            assert not self.left_data
            # TODO: special case this code path
            partId = self.nextPartId if leftPartId is None else leftPartId
            spill_data = self.data
            self.parts.append(write(partId, spill_data))
            self.data = bytearray()

            if finalise:
                rr = write.finalise(self.parts)

            return len(spill_data), rr

        bytes_written = 0
        if self.data:
            self.is_final = True
            bytes_written = self.flush_rhs(write)

        if self.left_data:
            assert len(self.left_data) >= write.min_write_sz
            partId = 1 if leftPartId is None else leftPartId
            self.parts.insert(0, write(partId, self.left_data))
            bytes_written += len(self.left_data)
            self.left_data = bytearray()

        if finalise:
            rr = write.finalise(self.parts)

        return bytes_written, rr

    def maybe_write(self, write: PartsWriter, spill_sz: int) -> int:
        # if not last section keep 'min_write_sz' and 1 partId around after flush
        rhs_keep, parts_to_keep = (0, 0) if self.is_final else (write.min_write_sz, 1)
        lhs_keep = 0 if self.started_write else self.lhs_keep

        if self.write_credits - 1 < parts_to_keep:
            return 0

        bytes_to_write = len(self.data) - rhs_keep - lhs_keep
        if bytes_to_write < spill_sz:
            return 0

        if lhs_keep == 0:
            spill_data = self.data[:bytes_to_write]
            self.data = bytearray(self.data[bytes_to_write:])
        else:
            spill_data = self.data[lhs_keep : lhs_keep + bytes_to_write]
            assert not self.left_data
            self.left_data = bytearray(self.data[:lhs_keep])
            self.data = bytearray(self.data[bytes_to_write + lhs_keep :])

        assert len(spill_data) == bytes_to_write
        assert len(spill_data) >= spill_sz

        self.parts.append(write(self.nextPartId, spill_data))
        self.nextPartId += 1
        self.write_credits -= 1

        return bytes_to_write

    @staticmethod
    def gen_bunch(
        partId: int,
        n: int,
        *,
        writes_per_chunk: int = 1,
        mark_final: bool = False,
        lhs_keep: int = 0,
    ) -> Iterator["MPUChunk"]:
        for idx in range(n):
            is_final = mark_final and idx == (n - 1)
            yield MPUChunk(
                partId + idx * writes_per_chunk,
                writes_per_chunk,
                is_final=is_final,
                lhs_keep=lhs_keep,
            )

    @staticmethod
    def from_dask_bag(
        partId: int,
        chunks: "dask.bag.Bag",
        *,
        writes_per_chunk: int = 1,
        mark_final: bool = False,
        lhs_keep: int = 0,
        write: Optional[PartsWriter] = None,
        spill_sz: int = 0,
        split_every: int = 4,  # when applying fold take 4 at a time
    ) -> "dask.bag.Item":
        # pylint: disable=import-outside-toplevel
        import dask.bag

        mpus = dask.bag.from_sequence(
            MPUChunk.gen_bunch(
                partId,
                chunks.npartitions,
                writes_per_chunk=writes_per_chunk,
                mark_final=mark_final,
                lhs_keep=lhs_keep,
            ),
            npartitions=chunks.npartitions,
        )

        mpus = dask.bag.map_partitions(
            _mpu_append_chunks_op,
            mpus,
            chunks,
            write=write,
            spill_sz=spill_sz,
            token="mpu.append",
        )

        return mpus.fold(
            partial(_merge_and_spill_op, write=write, spill_sz=spill_sz),
            split_every=split_every,
        )

    @staticmethod
    def collate_substreams(
        substreams: list["dask.bag.Item"],
        *,
        write: Optional[PartsWriter] = None,
        spill_sz: int = 0,
    ) -> "dask.bag.Item":
        # pylint: disable=import-outside-toplevel
        import dask.bag
        from dask import delayed

        assert len(substreams) > 0

        return dask.bag.Item.from_delayed(
            delayed(_mpu_collate_op)(
                substreams, pure=False, write=write, spill_sz=spill_sz
            )
        )


def mpu_write(
    chunks: "dask.bag.Bag" | list["dask.bag.Bag"],
    write: PartsWriter | None = None,
    *,
    mk_header: Any = None,
    mk_footer: Any = None,
    user_kw: dict[str, Any] | None = None,
    writes_per_chunk: int = 1,
    spill_sz: int = 20 * (1 << 20),
    dask_name_prefix="mpufinalise",
) -> "Delayed":
    # pylint: disable=import-outside-toplevel,too-many-locals,too-many-arguments
    from dask.base import tokenize
    from dask.delayed import delayed

    if not isinstance(chunks, list):
        chunks = [chunks]
    if write is None:
        min_part = 1
        lhs_keep = 0
    else:
        min_part = write.min_part
        lhs_keep = write.min_write_sz

    partId = min_part + 1
    dss: list["dask.bag.Item"] = []
    for idx, ch in enumerate(chunks):
        sub = MPUChunk.from_dask_bag(
            partId,
            ch,
            writes_per_chunk=writes_per_chunk,
            lhs_keep=lhs_keep,
            spill_sz=spill_sz,
            mark_final=mk_footer is None and (idx == len(chunks) - 1),
            write=write,
        )
        dss.append(sub)
        partId = partId + ch.npartitions * writes_per_chunk

    if len(dss) == 1:
        data_substream = dss[0]
    else:
        data_substream = MPUChunk.collate_substreams(
            dss,
            write=write,
            spill_sz=spill_sz,
        )

    tk = tokenize(write, mk_header, mk_footer, user_kw, spill_sz)
    name = f"{dask_name_prefix}-{tk}"

    return delayed(_finalizer_dask_op, name=name, pure=True)(
        data_substream,
        write=write,
        mk_header=mk_header,
        mk_footer=mk_footer,
        user_kw=user_kw,
        dask_key_name=name,
    )


def _mpu_collate_op(
    substreams: list[MPUChunk],
    *,
    write: Optional[PartsWriter] = None,
    spill_sz: int = 0,
) -> MPUChunk:
    assert len(substreams) > 0
    root, *rest = substreams
    for rhs in rest:
        root = MPUChunk.merge(root, rhs, write=write)
        if write and spill_sz:
            root.maybe_write(write, spill_sz)
    return root


def _mpu_append_chunks_op(
    mpus: Iterable[MPUChunk],
    chunks: Iterable[tuple[bytes, Any]],
    write: Optional[PartsWriter] = None,
    spill_sz: int = 0,
):
    # expect 1 MPUChunk per partition
    (mpu,) = mpus
    for chunk in chunks:
        data, chunk_id = chunk
        mpu.append(data, chunk_id)
        if write is not None and spill_sz > 0:
            mpu.maybe_write(write, spill_sz)

    return [mpu]


def _merge_and_spill_op(
    lhs: MPUChunk,
    rhs: MPUChunk,
    write: Optional[PartsWriter] = None,
    spill_sz: int = 0,
) -> MPUChunk:
    mm = MPUChunk.merge(lhs, rhs, write)
    if write is None or spill_sz == 0:
        return mm

    mm.maybe_write(write, spill_sz)
    return mm


def _finalizer_dask_op(
    data_substream: MPUChunk,
    *,
    write: PartsWriter | None = None,
    mk_header: Any = None,
    mk_footer: Any = None,
    user_kw: dict[str, Any] | None = None,
):
    if user_kw is None:
        user_kw = {}

    _root = data_substream
    hdr_bytes, footer_bytes = [
        None if op is None else op(data_substream.observed, **user_kw)
        for op in [mk_header, mk_footer]
    ]

    if footer_bytes:
        _root.append(footer_bytes)

    if hdr_bytes:
        hdr = MPUChunk(1, 1)
        hdr.append(hdr_bytes)
        _root = MPUChunk.merge(hdr, _root)

    if write is None:
        return _root

    _, rr = _root.flush(write, leftPartId=1, finalise=True)
    return rr


def get_mpu_kwargs(
    mk_header=None,
    mk_footer=None,
    user_kw=None,
    writes_per_chunk=1,
    spill_sz=20 * (1 << 20),
    client=None,
) -> dict:
    """
    Construct shared keyword arguments for multipart uploads.
    """
    return {
        "mk_header": mk_header,
        "mk_footer": mk_footer,
        "user_kw": user_kw,
        "writes_per_chunk": writes_per_chunk,
        "spill_sz": spill_sz,
        "client": client,
    }


def mpu_upload(
    chunks: Union[dask.bag.Bag, list[dask.bag.Bag]],
    *,
    writer: Any,
    dask_name_prefix: str,
    **kw,
) -> "Delayed":
    """Shared logic for multipart uploads to storage services."""
    client = kw.pop("client", None)
    writer_kw = dict(kw)
    if client is not None:
        writer_kw["client"] = client
    spill_sz = kw.get("spill_sz", 20 * (1 << 20))
    if spill_sz:
        write = writer(writer_kw)
    else:
        write = None
    return mpu_write(
        chunks,
        write,
        dask_name_prefix=dask_name_prefix,
        **kw,  # everything else remains
    )



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import importlib
import itertools
import math
from collections import OrderedDict, namedtuple
from typing import (
    Any,
    Callable,
    Dict,
    Iterator,
    List,
    Literal,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import numpy
from affine import Affine

from . import geom
from .crs import CRS, MaybeCRS, SomeCRS, norm_crs
from .geom import BoundingBox, Geometry, bbox_intersection, bbox_union
from .math import (
    clamp,
    extract_anchor,
    is_affine_st,
    is_almost_int,
    maybe_zero,
    resolution_from_affine,
    snap_affine,
    snap_grid,
    split_translation,
)
from .roi import (
    RoiTiles,
    align_up,
    clip_tiles,
    roi_boundary,
    roi_normalise,
    roi_shape,
    roi_tiles,
)
from .types import (
    ROI,
    XY,
    AnchorEnum,
    Chunks2d,
    MaybeInt,
    NormalizedROI,
    OutlineMode,
    Resolution,
    Shape2d,
    SomeIndex2d,
    SomeResolution,
    SomeShape,
    Unset,
    func2map,
    res_,
    shape_,
    xy_,
)

GeoboxAnchor = Union[
    AnchorEnum,
    XY[float],
    float,
    Literal["center"],
    Literal["centre"],
    Literal["edge"],
    Literal["floating"],
    Literal["default"],
]

# pylint: disable=invalid-name,too-many-public-methods,too-many-lines
Coordinate = namedtuple("Coordinate", ("values", "units", "resolution"))


def _norm_anchor(anchor: GeoboxAnchor) -> Union[AnchorEnum, XY[float]]:
    if isinstance(anchor, AnchorEnum):
        return anchor
    if isinstance(anchor, (float, int)):
        if anchor == 0:
            return AnchorEnum.EDGE
        if anchor == 0.5:
            return AnchorEnum.CENTER
        return xy_(float(anchor), float(anchor))
    if isinstance(anchor, XY):
        return anchor

    return {
        "center": AnchorEnum.CENTER,
        "centre": AnchorEnum.CENTER,
        "edge": AnchorEnum.EDGE,
        "floating": AnchorEnum.FLOATING,
        "default": AnchorEnum.EDGE,
    }[anchor]


class GeoBoxBase:
    """
    Defines the location and resolution of a rectangular grid of data,
    including it's :py:class:`~odc.geo.crs.CRS`.

    :param shape: Shape in pixels ``(ny, nx)``
    :param crs: Coordinate Reference System
    :param affine: Affine transformation defining the location of the geobox
    """

    __slots__ = ("_shape", "_affine", "_crs", "_extent", "_lazy_ui")

    def __init__(self, shape: SomeShape, affine: Affine, crs: MaybeCRS):
        shape = shape_(shape)

        self._shape = shape
        self._affine = affine
        self._crs = norm_crs(crs)
        self._extent: Optional[Geometry] = None
        self._lazy_ui = None

        if self.shape.x < 0 or self.shape.y < 0:
            raise ValueError(
                f"Got shape {self._shape.yx!r}: negative sizes are not allowed."
            )

    @property
    def width(self) -> int:
        """Width in pixels (nx)."""
        return self._shape.x

    @property
    def height(self) -> int:
        """Height in pixels (ny)."""
        return self._shape.y

    @property
    def shape(self) -> Shape2d:
        """Shape in pixels ``(height, width)``."""
        return self._shape

    @property
    def aspect(self) -> float:
        """Aspect ratio (X/Y in pixel space)."""
        return self._shape.aspect

    @property
    def crs(self) -> Optional[CRS]:
        """Coordinate Reference System of the GeoBox."""
        return self._crs

    @property
    def dimensions(self) -> Tuple[str, str]:
        """List of dimension names of the GeoBox."""
        crs = self._crs
        if crs is None:
            return ("y", "x")
        return crs.dimensions

    dims = dimensions

    def is_empty(self) -> bool:
        """Check if geobox is "empty"."""
        return 0 in self._shape

    def __bool__(self) -> bool:
        return not self.is_empty()

    @property
    def resolution(self) -> Resolution:
        """Resolution, pixel size in CRS units."""
        return resolution_from_affine(self._affine)

    def boundary(self, pts_per_side: int = 16) -> numpy.ndarray:
        """
        Boundary of a :py:class:`~odc.geo.geobox.GeoBox`.

        Construct a ring of points in pixel space along the edge of the geobox.

        :param pts_per_side: Number of points per side, default is 16.

        :return:
          Points in pixel space along the perimeter of a GeoBox as a ``Nx2`` array
          in pixel coordinates.
        """
        ny, nx = self._shape.yx
        return roi_boundary(numpy.s_[0:ny, 0:nx], pts_per_side)

    @property
    def alignment(self) -> XY[float]:
        """
        Alignment of pixel boundaries in CRS units.

        This is usally ``(0,0)``.

        .. seealso:: :py:meth:`~odc.geo.geobox.GeoBoxBase.anchor`
        """
        # TODO: deprecate alignment method
        rx, _, tx, _, ry, ty, *_ = self._affine
        return xy_(tx % abs(rx), ty % abs(ry))

    @property
    def anchor(self) -> XY[float] | AnchorEnum:
        """
        Anchor point in pixel coordinates.

        This is only valid for linear sources.
        """
        if not self.linear:
            return AnchorEnum.FLOATING
        return extract_anchor(self._affine)

    @property
    def linear(self) -> bool:
        return True

    def wld2pix(self, x, y):
        return (~self._affine) * (x, y)

    def pix2wld(self, x, y):
        return self._affine * (x, y)

    @property
    def extent(self) -> Geometry:
        """GeoBox footprint in native CRS."""
        if self._extent is not None:
            return self._extent

        if self.linear:
            _extent = geom.polygon_from_transform(self._shape, self._affine, self._crs)
        else:
            _extent = geom.polygon(self.boundary(16).tolist(), self._crs).transform(
                self.pix2wld
            )

        self._extent = _extent
        return _extent

    @property
    def boundingbox(self) -> BoundingBox:
        """GeoBox bounding box in the native CRS."""
        return BoundingBox.from_transform(self._shape, self._affine, crs=self._crs)

    def _reproject_resolution(self, npoints: int = 100):
        bbox = self.extent.boundingbox
        span = max(bbox.span_x, bbox.span_y)
        return span / npoints

    def footprint(
        self, crs: SomeCRS, buffer: float = 0, npoints: int = 100
    ) -> Geometry:
        """
        Compute footprint in foreign CRS.

        :param crs: CRS of the destination
        :param buffer: amount to buffer in source pixels before transforming
        :param npoints: number of points per-side to use, higher number
                        is slower but more accurate
        """
        assert self.crs is not None
        ext = self.extent
        if buffer != 0:
            buffer = buffer * max(*self.resolution.xy)
            ext = ext.buffer(buffer)

        return ext.to_crs(crs, resolution=self._reproject_resolution(npoints)).dropna()

    @property
    def geographic_extent(self) -> Geometry:
        """GeoBox extent in EPSG:4326."""
        if self._crs is None or self._crs.geographic:
            return self.extent
        return self.footprint("epsg:4326")

    @property
    def _ui(self):
        # pylint: disable=import-outside-toplevel
        from .ui import PixelGridDisplay

        if self._lazy_ui is not None:
            return self._lazy_ui

        gsd = max(*self.resolution.map(abs).xy)
        self._lazy_ui = PixelGridDisplay(self, self.pix2wld, gsd)
        return self._lazy_ui

    def svg(
        self,
        scale_factor: float = 1.0,
        mode: OutlineMode = "auto",
        notch: float = 0.0,
        grid_stroke: str = "pink",
    ) -> str:
        """
        Produce SVG paths.

        :param mode: One of pixel, native, geo (default is geo)
        :return: SVG path
        """
        return self._ui.svg(
            scale_factor=scale_factor, mode=mode, notch=notch, grid_stroke=grid_stroke
        )

    def grid_lines(self, step: int = 0, mode: OutlineMode = "native") -> Geometry:
        """
        Construct pixel edge aligned grid lines.
        """
        return self._ui.grid_lines(step=step, mode=mode)

    def outline(self, mode: OutlineMode = "native", notch: float = 0.1) -> Geometry:
        return self._ui.outline(mode, notch=notch)

    def _repr_svg_(self):
        # pylint: disable=protected-access
        return self._ui._render_svg()

    def _repr_html_(self):
        # pylint: disable=protected-access
        return self._ui._repr_html_()

    def compute_crop(self, roi) -> Tuple[Shape2d, Affine]:
        if isinstance(roi, BoundingBox):
            roi = roi.polygon
        if isinstance(roi, GeoBoxBase):
            roi = roi.extent

        if isinstance(roi, Geometry):
            if roi.crs is not None:
                roi = self.project(roi)
            pix_bbox = roi.boundingbox.round() & BoundingBox(
                0, 0, self.width, self.height
            )
            nx, ny = (max(1, int(span)) for span in (pix_bbox.span_x, pix_bbox.span_y))
            tx, ty = map(int, pix_bbox.bbox[:2])
            roi = numpy.s_[ty : ty + ny, tx : tx + nx]

        if isinstance(roi, int):
            roi = (slice(roi, roi + 1), slice(None, None))

        if isinstance(roi, slice):
            roi = (roi, slice(None, None))

        if len(roi) > 2:
            raise ValueError("Expect 2d slice")

        roi = roi_normalise(roi, self._shape.shape)

        if not all(s.step is None or s.step == 1 for s in roi):
            raise NotImplementedError("scaling not implemented, yet")

        ty, tx = (s.start for s in roi)
        ny, nx = roi_shape(roi)

        affine = self._affine * Affine.translation(tx, ty)
        return shape_((ny, nx)), affine

    def compute_zoom_out(self, factor: float) -> Tuple[Shape2d, Affine]:
        ny, nx = (max(1, math.ceil(s / factor)) for s in self.shape)
        A = self._affine * Affine.scale(factor, factor)
        return (shape_((ny, nx)), A)

    def compute_zoom_to(
        self,
        shape: Union[SomeShape, int, float, None] = None,
        *,
        resolution: Optional[SomeResolution] = None,
    ) -> Tuple[Shape2d, Affine]:
        """
        Change GeoBox shape.

        When supplied a single integer scale longest dimension to match that.

        :returns:
          GeoBox covering the same region but with different number of pixels and therefore resolution.
        """
        if shape is None:
            if resolution is None:
                raise ValueError("Have to supply shape or resolution")
            new_geobox = GeoBox.from_bbox(
                self.boundingbox, resolution=resolution, tight=True
            )
            return new_geobox.shape, new_geobox.affine

        if isinstance(shape, (int, float)):
            nmax = max(*self._shape)
            return self.compute_zoom_out(nmax / shape)

        shape = shape_(shape)
        sy, sx = (N / float(n) for N, n in zip(self._shape, shape.shape))
        A = self._affine * Affine.scale(sx, sy)
        return (shape, A)

    def project(self, g: Geometry) -> Geometry:
        """
        Map Geometry between world and pixel coords.

        When input geometry has no CRS, map from pixels to the world.

        When input geometry has CRS (can be different from GeoBox), project
        geometry into pixel coordinates, note that result is not clipped to the
        image bounds.
        """
        if g.crs is None:  # assume pixel plane
            g = g.transform(self.pix2wld)
            return Geometry(g.geom, self._crs)

        assert self._crs is not None
        if g.crs != self._crs:
            g = g.to_crs(self._crs)
        g = g.transform(self.wld2pix)
        return Geometry(g.geom, crs=None)

    def qr2sample(
        self,
        n: int,
        padding: Optional[float] = None,
        with_edges: bool = False,
        offset: int = 0,
    ) -> Geometry:
        """
        Generate quasi-random sample of image locations.

        :param n:
           Number of points

        :param padding:
           In pixels, minimal distance from the edge

        :param offset:
           Offset into quasi-random sequence from where to start

        :param edges:
           Also include samples along the edge and corners

        References: http://extremelearning.com.au/unreasonable-effectiveness-of-quasirandom-sequences/
        """
        nx, ny = self._shape.xy
        return BoundingBox(0, 0, nx, ny, None).qr2sample(
            n, padding=padding, with_edges=with_edges, offset=offset
        )

    def explore(
        self,
        map: Optional[Any] = None,
        grid_lines: bool = True,
        tiles: Any = "OpenStreetMap",
        attr: Optional[str] = None,
        map_kwds: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Any:
        """
        Plot a visual representation of the GeoBox's extent on an
        interactive :py:mod:`folium` leaflet map.

        :param map:
            An optional existing :py:mod:`folium` map object to plot into.
            By default, a new map object will be created.
        :param: grid_lines:
            Whether to plot pixel edge aligned grid lines over the GeoBox
            extent.
        :param tiles:
            Map tileset to use for the map basemap. Supports any option
            supported by :py:mod:`folium`, including "OpenStreetMap",
            "CartoDB positron", "CartoDB dark_matter" or a custom XYZ URL.
        :param attr:
            Map tile attribution; only required if passing custom tile URL.
        :param map_kwds:
            Additional keyword arguments to pass to :py:class:`folium.Map`.
        :param kwargs:
            Additional keyword arguments to pass to :py:class:`folium.GeoJson`.

        :return: A :py:mod:`folium` map containing the plotted GeoBox.
        """
        # pylint: disable=redefined-builtin

        # Add outline to map
        map = self.outline(mode="geo", notch=0.0).explore(
            map=map, tiles=tiles, attr=attr, map_kwds=map_kwds, **kwargs
        )

        # Optionally overlay grid lines
        if grid_lines:
            self.grid_lines(mode="geo").explore(
                map=map,
                style_function=lambda feature: {
                    "weight": 1,
                    "fillOpacity": 0.3,
                },
            )
        return map

    def __getitem__(self, roi) -> "GeoBoxBase":
        raise NotImplementedError()


class GeoBox(GeoBoxBase):
    """
    Defines the location and resolution of a rectangular grid of data,
    including it's :py:class:`~odc.geo.crs.CRS`.

    :param shape: Shape in pixels ``(ny, nx)``
    :param crs: Coordinate Reference System
    :param affine: Affine transformation defining the location of the geobox
    """

    __slots__ = ()

    def __init__(self, shape: SomeShape, affine: Affine, crs: MaybeCRS):
        GeoBoxBase.__init__(self, shape, affine, crs)

    @staticmethod
    def from_bbox(
        bbox: Union[BoundingBox, Tuple[float, float, float, float]],
        crs: MaybeCRS = None,
        *,
        tight: bool = False,
        shape: Union[SomeShape, int, None] = None,
        resolution: Optional[SomeResolution] = None,
        anchor: GeoboxAnchor = "default",
        tol: float = 0.01,
    ) -> "GeoBox":
        """
        Construct :py:class:`~odc.geo.geobox.GeoBox` from a bounding box.

        :param bbox: Bounding box in CRS units, lonlat is assumed when ``crs`` is not supplied
        :param crs: CRS of the bounding box (defaults to EPSG:4326)
        :param shape:
           Span that many pixels, if it's a single number then span that many pixels along the
           longest dimension, other dimension will be computed to maintain roughly square pixels.
        :param resolution: Use specified resolution
        :param tight: Supplying ``tight=True`` turns off pixel snapping.
        :param anchor:
            By default snaps grid such that pixel edges fall on X/Y axis. Ignored when tight mode is
            used.
        :param tol:
            Fraction of a pixel that can be ignored, defaults to 1/100. Bounding box of the output
            geobox is allowed to be smaller than supplied bounding box by that amount.

        :return:
           :py:class:`~odc.geo.geobox.GeoBox` that covers supplied bounding box.
        """
        # pylint: disable=too-many-locals, too-many-branches

        anchor = _norm_anchor(anchor)
        _snap: Optional[XY[float]] = None

        if tight:
            anchor = AnchorEnum.FLOATING

        if isinstance(anchor, XY):
            _snap = anchor
        if anchor == AnchorEnum.EDGE:
            _snap = xy_(0, 0)
        elif anchor == AnchorEnum.CENTER:
            _snap = xy_(0.5, 0.5)

        def _norm_bbox(
            bbox: Tuple[float, float, float, float], crs: MaybeCRS
        ) -> BoundingBox:
            if isinstance(crs, str):
                if crs.lower().startswith("utm"):
                    return BoundingBox(*bbox, crs="epsg:4326").to_crs(crs)

            return BoundingBox(*bbox, crs=(crs or "epsg:4326"))

        if not isinstance(bbox, BoundingBox):
            bbox = _norm_bbox(bbox, crs)
        elif bbox.crs is None:
            bbox = _norm_bbox(bbox.bbox, crs)

        if isinstance(shape, (int, float)):
            if bbox.aspect > 1:
                resolution = bbox.span_x / shape
            else:
                resolution = bbox.span_y / shape
            shape = None

        if resolution is not None:
            rx, ry = res_(resolution).xy
            if _snap is None:
                offx, nx = snap_grid(bbox.left, bbox.right, rx, None, tol=tol)
                offy, ny = snap_grid(bbox.bottom, bbox.top, ry, None, tol=tol)
            else:
                offx, nx = snap_grid(bbox.left, bbox.right, rx, _snap.x, tol=tol)
                offy, ny = snap_grid(bbox.bottom, bbox.top, ry, _snap.y, tol=tol)

            affine = Affine.translation(offx, offy) * Affine.scale(rx, ry)
            return GeoBox((ny, nx), crs=bbox.crs, affine=affine)

        if shape is None:
            raise ValueError("Must supply shape or resolution")

        shape = shape_(shape)
        nx, ny = shape.wh

        if nx == 0 or ny == 0:
            raise ValueError(
                f"Got zero size in ({nx}, {ny}): cannot calculate resolution."
            )

        if nx < 0 or ny < 0:
            raise ValueError(
                f"Got negative size in ({nx}, {ny}): negative dimensions are not allowed."
            )

        rx = bbox.span_x / nx
        ry = -bbox.span_y / ny

        if _snap is None:
            offx, offy = bbox.left, bbox.top
        else:
            offx, _ = snap_grid(bbox.left, bbox.right, rx, _snap.x, tol=tol)
            offy, _ = snap_grid(bbox.bottom, bbox.top, ry, _snap.y, tol=tol)

        affine = Affine.translation(offx, offy) * Affine.scale(rx, ry)
        return GeoBox((ny, nx), crs=bbox.crs, affine=affine)

    @staticmethod
    def from_geopolygon(
        geopolygon: Geometry,
        resolution: Optional[SomeResolution] = None,
        crs: MaybeCRS = None,
        align: Optional[XY[float]] = None,
        *,
        shape: Union[SomeShape, int, None] = None,
        tight: bool = False,
        anchor: GeoboxAnchor = "default",
        tol: float = 0.01,
    ) -> "GeoBox":
        """
        Construct :py:class:`~odc.geo.geobox.GeoBox` from a polygon.

        :param resolution:
           Either a single number or a :py:class:`~odc.geo.types.Resolution` object.

        :param shape:
           Span that many pixels, if it's a single number then span that many pixels along the
           longest dimension, other dimension will be computed to maintain roughly square pixels.

        :param crs:
           CRS to use, if different from the geopolygon

        :param align:
            Deprecated: please switch to ``anchor=``

        :param anchor:
            By default snaps grid such that pixel edges fall on X/Y axis.

        :param tol:
            Fraction of a pixel that can be ignored, defaults to 1/100. Bounding box of the output
            geobox is allowed to be smaller than supplied bounding box by that amount.

        :param tight: Supplying ``tight=True`` turns off pixel snapping.

        """
        if align is not None:
            # support old-style "align", which is basically anchor but in CRS units
            ax, ay = align.xy
            if ax == 0 and ay == 0:
                anchor = AnchorEnum.EDGE
            else:
                assert resolution is not None
                resolution = res_(resolution)
                anchor = xy_(ax / abs(resolution.x), ay / abs(resolution.y))

        if crs is None or isinstance(crs, Unset):
            crs = geopolygon.crs
        else:
            geopolygon = geopolygon.to_crs(crs)

        return GeoBox.from_bbox(
            geopolygon.boundingbox,
            crs,
            shape=shape,
            resolution=resolution,
            anchor=anchor,
            tol=tol,
            tight=tight,
        )

    @staticmethod
    def from_rio(rdr) -> "GeoBox":
        """
        Construct GeoBox from rasterio.

        :param rdr: Openned :py:class:`rasterio.DatasetReader`
        :returns:
           :py:class:`~odc.geo.geobox.GeoBox`
        """
        return GeoBox(rdr.shape, rdr.transform, rdr.crs)

    def buffered(self, xbuff: float, ybuff: Optional[float] = None) -> "GeoBox":
        """
        Produce a tile buffered by ``xbuff, ybuff`` (in CRS units).
        """
        if ybuff is None:
            ybuff = xbuff

        by, bx = (
            _round_to_res(buf, res)
            for buf, res in zip((ybuff, xbuff), self.resolution.yx)
        )
        affine = self._affine * Affine.translation(-bx, -by)

        ny, nx = (sz + 2 * b for sz, b in zip(self._shape, (by, bx)))

        return GeoBox(
            (ny, nx),
            affine=affine,
            crs=self._crs,
        )

    def enclosing(self, region: Union[Geometry, BoundingBox]) -> "GeoBox":
        """
        Construct compatible geobox covering given ``region``.

        Output GeoBox shares exactly the same pixel grid as source, but has
        different shape and location in the world.

        :param region: Region to be covered by the new GeoBox.
        """
        if isinstance(region, BoundingBox):
            region = region.polygon

        if region.crs is None:
            raise ValueError("Must supply geo-resgistered region")

        pix_bbox = self.project(region).boundingbox.round()
        nx, ny = (max(1, int(span)) for span in (pix_bbox.span_x, pix_bbox.span_y))
        tx, ty, *_ = pix_bbox.bbox
        A = self.translate_pix(tx, ty).affine

        return GeoBox(shape_((ny, nx)), A, self._crs)

    def __getitem__(self, roi) -> "GeoBox":
        _shape, _affine = self.compute_crop(roi)
        return GeoBox(shape=_shape, affine=_affine, crs=self._crs)

    def __or__(self, other) -> "GeoBox":
        """A geobox that encompasses both self and other."""
        return geobox_union_conservative([self, other])

    def __and__(self, other) -> "GeoBox":
        """A geobox that is contained in both self and other."""
        return geobox_intersection_conservative([self, other])

    def __hash__(self):
        return hash((*self._shape, self._crs, self._affine))

    def overlap_roi(self, other: "GeoBox", tol: float = 1e-8) -> NormalizedROI:
        """
        Compute overlap as ROI.

        Figure out slice into this geobox that shares pixels with the ``other`` geobox with
        consistent pixel grid.

        :raises:
            :py:class:`ValueError` when two geoboxes are not pixel-aligned.
        """
        nx, ny = self._shape.xy
        x0, y0, x1, y1 = map(int, bounding_box_in_pixel_domain(other, self, tol))
        x0, y0 = max(0, x0), max(0, y0)
        x1, y1 = min(x1, nx), min(y1, ny)
        return numpy.s_[y0:y1, x0:x1]

    @property
    def transform(self) -> Affine:
        """Linear mapping from pixel space to CRS."""
        return self._affine

    @property
    def affine(self) -> Affine:
        """
        Linear mapping from pixel space to CRS.

        alias for :py:attr:`~odc.geo.geobox.GeoBox.transform`
        """
        return self._affine

    @property
    def coordinates(self) -> Dict[str, Coordinate]:
        """
        Query coordinates.

        This method only works with axis-aligned boxes. It will raise :py:class:`ValueError` if called
        on non-axis aligned :py:class:`~odc.geo.geobox.GeoBox`.

        :raises: :py:class:`ValueError` if not axis aligned.

        :return:
          Mapping from coordinate name to :py:class:`~odc.geo.geobox.Coordinate`.
        """
        self._confirm_axis_aligned("Only axis aligned GeoBox can do this.")
        rx, _, tx, _, ry, ty, *_ = self._affine
        ny, nx = self._shape

        xs = numpy.arange(nx) * rx + (tx + rx / 2)
        ys = numpy.arange(ny) * ry + (ty + ry / 2)

        crs_units = self._crs.units if self._crs is not None else ("1", "1")

        return OrderedDict(
            (dim, Coordinate(labels, units, res))
            for dim, labels, units, res in zip(
                self.dimensions, (ys, xs), crs_units, (ry, rx)
            )
        )

    coords = coordinates

    def map_bounds(self) -> Tuple[Tuple[float, float], Tuple[float, float]]:
        """
        Query bounds in folium/ipyleaflet style.

        Returns SW, and NE corners in lat/lon order.
        ``((lat_w, lon_s), (lat_e, lon_n))``.
        """
        if self._crs is not None:
            (x0, y0), _, (x1, y1) = self.extent.exterior.to_crs("epsg:4326").points[:3]
        else:
            (x0, y0), _, (x1, y1) = self.extent.exterior.points[:3]
        return (y0, x0), (y1, x1)

    def to_crs(
        self,
        crs: SomeCRS,
        *,
        resolution: Literal["auto", "fit", "same"] = "auto",
        shape: Union[SomeShape, int, None] = None,
        tight: bool = False,
        anchor: GeoboxAnchor = "default",
        tol: float = 0.01,
        round_resolution: Union[None, bool, Callable[[float, str], float]] = None,
    ) -> "GeoBox":
        """
        Compute GeoBox covering the same region in a different projection.

        :param crs:
           Desired CRS of the output

        :param resolution:

           * "same" use exactly the same resolution as src
           * "fit" use center pixel to determine scale change between the two
           * | "auto" is to use the same resolution on the output if CRS units are the same
             |  between the source and destination and otherwise use "fit"

        :param shape:
          Span that many pixels, if it's a single number then span that many pixels
          along the longest dimension, other dimension will be computed to maintain
          roughly square pixels. Takes precedence over ``resolution=`` parameter.

        :param tight:
          By default output pixel grid is adjusted to align pixel edges to X/Y
          axis, suppling ``tight=True`` produces unaligned geobox on the output.

        :param anchor:
            Control pixel snapping, default is to snap pixel edge to
            ``X=0,Y=0``. Ignored when ``tight=True`` is supplied.

        :param tol:
            Fraction of the output pixel that can be ignored, defaults to 1/100.
            Bounding box of the output geobox is allowed to be smaller by that
            amount than transformed footprint of the original.

        :param round_resolution:
          ``round_resolution(res: float, units: str) -> float``

        :return:
           Similar resolution, axis aligned geobox that fully encloses this one
           but in a different projection.
        """
        # pylint: disable=import-outside-toplevel,too-many-arguments
        # can't be up-top due to circular imports issues
        from .overlap import compute_output_geobox

        return compute_output_geobox(
            self,
            crs,
            resolution=resolution,
            shape=shape,
            tight=tight,
            anchor=anchor,
            round_resolution=round_resolution,
            tol=tol,
        )

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return f"GeoBox({self._shape.yx!r}, {self._affine!r}, {self._crs!r})"

    def __eq__(self, other):
        if not isinstance(other, GeoBox):
            return False

        return (
            self._shape == other._shape
            and self._affine == other._affine
            and self._crs == other._crs
        )

    def __rmul__(self, transform: Affine) -> "GeoBox":
        """
        Apply affine transform on CRS side.

        This has effect of transforming footprint of the source via ``transform``.

        :param transform:
           Affine matrix that shifts footprint of the source geobox.

        :return:
           :py:class:`~odc.geo.gebox.GeoBox` of the same pixel shape but
           covering different region.
        """
        return GeoBox(self._shape, transform * self._affine, self._crs)

    def __mul__(self, transform: Affine) -> "GeoBox":
        """
        Apply affine transform on pixel side.

        ``X_old_pix = transform * X_new_pix``

        :param transform:
           Affine matrix mapping from new pixel coordinate space to pixel coordinate
           space of input geobox.

        :returns:
          :py:class:`~odc.geo.gebox.GeoBox` of the same pixel shape but covering different
          region. Pixel coordinates in the output relate to input coordinates via ``transform``.
        """
        return GeoBox(self._shape, self._affine * transform, self._crs)

    def snap_to(self, other: "GeoBox") -> "GeoBox":
        """
        Snap pixel grid to ``other``.

        Find smallest sub-pixel translation to apply to this geobox such that
        pixel edges align with ``other``.

        :param other: GeoBox to snap to, must be related by translation only, no
            change in scale or rotation.

        :raises: ``ValueError`` when ``other`` is in a different projection or
            has different resolution or orientation.
        """
        _, subpix = split_translation(pixel_translation(other, self))
        tx, ty = subpix.map(lambda x: maybe_zero(x, 1e-8)).xy
        return self.translate_pix(tx, ty)

    def pad(self, padx: int, pady: MaybeInt = None) -> "GeoBox":
        """
        Pad geobox.

        Expand GeoBox by fixed number of pixels on each side
        """
        # false positive for -pady, it's never None by the time it runs
        # pylint: disable=invalid-unary-operand-type

        pady = padx if pady is None else pady

        ny, nx = self._shape.yx
        A = self._affine * Affine.translation(-padx, -pady)
        shape = (ny + pady * 2, nx + padx * 2)
        return GeoBox(shape, A, self._crs)

    def pad_wh(self, alignx: int = 16, aligny: MaybeInt = None) -> "GeoBox":
        """
        Possibly expand :py:class:`~odc.geo.geobox.GeoBox` by a few pixels.

        Find nearest ``width``/``height`` that are multiples of the desired factor. And return a new
        geobox that is slighly taller and/or wider covering roughly the same region. The new geobox
        will have the same CRS and transform but possibly larger shape.
        """
        aligny = alignx if aligny is None else aligny
        ny, nx = (align_up(sz, n) for sz, n in zip(self._shape.yx, (aligny, alignx)))

        return GeoBox((ny, nx), self._affine, self._crs)

    def crop(self, shape: SomeShape) -> "GeoBox":
        """
        Crop or expand to a given shape.

        :returns: New :py:class:`~odc.geo.geobox.GeoBox` with a new shape,
                 top left pixel remains at the same location and scale.
        """
        return GeoBox(shape, self._affine, self._crs)

    expand = crop

    def zoom_out(self, factor: float) -> "GeoBox":
        """
        Compute :py:class:`~odc.geo.geobox.GeoBox` with changed resolution.

        - ``factor > 1`` implies smaller width/height, fewer but bigger pixels
        - ``factor < 1`` implies bigger width/height, more but smaller pixels

        :returns:
           GeoBox covering the same region but with different pixels (i.e. lower or higher resolution)
        """
        _shape, _affine = self.compute_zoom_out(factor)
        return GeoBox(_shape, _affine, self._crs)

    def zoom_to(
        self,
        shape: Union[SomeShape, int, float, None] = None,
        *,
        resolution: Optional[SomeResolution] = None,
    ) -> "GeoBox":
        """
        Change GeoBox shape.

        When supplied a single integer scale longest dimension to match that.

        :returns:
          GeoBox covering the same region but with different number of pixels and therefore resolution.
        """
        _shape, _affine = self.compute_zoom_to(shape, resolution=resolution)
        return GeoBox(_shape, _affine, self._crs)

    def flipy(self) -> "GeoBox":
        """
        Flip along Y axis.

        :returns: GeoBox covering the same region but with Y-axis flipped
        """
        ny, _ = self._shape
        A = Affine.translation(0, ny) * Affine.scale(1, -1)
        return self * A

    def flipx(self) -> "GeoBox":
        """
        Flip along X axis.

        :returns: GeoBox covering the same region but with X-axis flipped
        """
        _, nx = self._shape
        A = Affine.translation(nx, 0) * Affine.scale(-1, 1)
        return self * A

    def translate_pix(self, tx: float, ty: float) -> "GeoBox":
        """
        Shift GeoBox in pixel plane.

        ``(0,0)`` of the new GeoBox will be at the same location as pixel ``(tx, ty)`` in the original
        GeoBox.
        """
        return self * Affine.translation(tx, ty)

    @property
    def left(self) -> "GeoBox":
        """Same size geobox to the left of this one."""
        return self.translate_pix(-self.shape.x, 0)

    @property
    def right(self) -> "GeoBox":
        """Same size geobox to the right of this one."""
        return self.translate_pix(self.shape.x, 0)

    @property
    def top(self) -> "GeoBox":
        """Same size geobox directly above this one."""
        return self.translate_pix(0, -self.shape.y)

    @property
    def bottom(self) -> "GeoBox":
        """Same size geobox directly below this one."""
        return self.translate_pix(0, self.shape.y)

    def rotate(self, deg: float) -> "GeoBox":
        """
        Rotate GeoBox around the center.

        It's as if you stick a needle through the center of the GeoBox footprint
        and rotate it counter clock wise by supplied number of degrees.

        Note that from the pixel point of view image rotates the other way. If you have
        source image with an arrow pointing right, and you rotate GeoBox 90 degrees,
        in that view arrow should point down (this is assuming usual case of inverted
        y-axis)
        """
        ny, nx = self._shape
        c0 = self._affine * (nx * 0.5, ny * 0.5)
        return Affine.rotation(deg, c0) * self

    def _confirm_axis_aligned(self, raise_error: Optional[str] = None) -> bool:
        if is_affine_st(self._affine):
            return True
        if raise_error is not None:
            raise ValueError(raise_error)
        return False

    @property
    def axis_aligned(self):
        """
        Check if Geobox is axis-aligned (not rotated).
        """
        return self._confirm_axis_aligned()

    @property
    def center_pixel(self) -> "GeoBox":
        """
        GeoBox of a center pixel.
        """
        return self[self.shape.map(lambda x: x // 2).yx]

    @property
    def compat(self):
        """
        Convert to :py:class:`datacube.utils.geometry.GeoBox`
        """
        try:
            dc_geom = importlib.import_module("datacube.utils.geometry")
        except ModuleNotFoundError:
            return None

        w, h = self.shape.wh
        return dc_geom.GeoBox(w, h, self._affine, str(self._crs))

    def __dask_tokenize__(self):
        return (
            "odc.geo.geobox.GeoBox",
            str(self.crs),
            *self._shape.yx,
            *self._affine[:6],
        )


def gbox_boundary(gbox: GeoBoxBase, pts_per_side: int = 16) -> numpy.ndarray:
    """Alias for :py:meth:`odc.geo.geobox.GeoBox.boundary`."""
    return gbox.boundary(pts_per_side)


def pixel_translation(a: GeoBox, b: GeoBox) -> XY[float]:
    """
    Compute pixel translation ``a -> b``.

    Both geoboxes should have the same CRS and resolution.
    """
    if a.crs != b.crs:
        raise ValueError("Geobox CRSs must match")

    # compute pixel-to-pixel transform
    # expect it to be a pure, pixel aligned translation
    #    1  0  tx
    #    0  1  ty
    #    0  0   1
    # Such that tx,ty are almost integer.
    sx, z1, tx, z2, sy, ty, *_ = ~b.affine * a.affine

    if not (
        numpy.isclose(sx, 1)
        and numpy.isclose(z1, 0)
        and numpy.isclose(z2, 0)
        and numpy.isclose(sy, 1)
    ):
        raise ValueError("Incompatible grids")

    return xy_(tx, ty)


def bounding_box_in_pixel_domain(
    geobox: GeoBox, reference: GeoBox, tol: float = 1e-8
) -> BoundingBox:
    """
    Bounding box of ``geobox`` in pixel space of ``reference``.

    :return:
      The bounding box of ``geobox`` with respect to the pixel grid defined by ``reference`` when
      their coordinate grids are compatible. Two geoboxes are compatible when they have the same
      CRS, same pixel size and orientation, and are related by whole pixel translation.

    :raises:
      :py:class:`ValueError` when two geoboxes are not pixel-aligned.
    """

    # offset of ``geobox`` in ``reference`` pixels
    tx, ty = pixel_translation(geobox, reference).xy

    if not (is_almost_int(tx, tol) and is_almost_int(ty, tol)):
        raise ValueError("Incompatible grids")

    tx, ty = round(tx), round(ty)
    ny, nx = geobox.shape
    return BoundingBox(tx, ty, tx + nx, ty + ny, None)


def geobox_union_conservative(geoboxes: List[GeoBox]) -> GeoBox:
    """
    Union of geoboxes as a geobox.

    Fails whenever incompatible grids are encountered.
    """
    if len(geoboxes) == 0:
        raise ValueError("No geoboxes supplied")

    reference, *_ = geoboxes

    bbox = bbox_union(
        bounding_box_in_pixel_domain(geobox, reference=reference) for geobox in geoboxes
    )

    affine = reference.affine * Affine.translation(*bbox[:2])
    return GeoBox(shape=bbox.shape, affine=affine, crs=reference.crs)


def geobox_intersection_conservative(geoboxes: List[GeoBox]) -> GeoBox:
    """
    Intersection of geoboxes.

    Fails whenever incompatible grids are encountered.
    """
    if len(geoboxes) == 0:
        raise ValueError("No geoboxes supplied")

    reference, *_ = geoboxes

    bbox = bbox_intersection(
        bounding_box_in_pixel_domain(geobox, reference=reference) for geobox in geoboxes
    )

    # standardise empty geobox representation
    if bbox.left > bbox.right:
        bbox = BoundingBox(
            left=bbox.left,
            bottom=bbox.bottom,
            right=bbox.left,
            top=bbox.top,
            crs=bbox.crs,
        )
    if bbox.bottom > bbox.top:
        bbox = BoundingBox(
            left=bbox.left,
            bottom=bbox.bottom,
            right=bbox.right,
            top=bbox.bottom,
            crs=bbox.crs,
        )

    affine = reference.affine * Affine.translation(*bbox[:2])

    return GeoBox(shape=bbox.shape, affine=affine, crs=reference.crs)


def scaled_down_geobox(src_geobox: GeoBox, scaler: int) -> GeoBox:
    """
    Compute :py:class:`~odc.geo.geobox.GeoBox` of a zoomed image.

    Given a source geobox and an integer scaler compute geobox of a scaled down image.

    Output geobox will be padded when shape is not a multiple of scaler.
    Example: ``5x4, scaler=2 -> 3x2``

    .. note::

       We assume that pixel coordinates are ``0,0`` at the top-left corner of a top-left pixel.

    """
    assert scaler > 1

    ny, nx = (X // scaler + (1 if X % scaler else 0) for X in src_geobox.shape)

    # Since 0,0 is at the corner of a pixel, not center, there is no
    # translation between pixel plane coords due to scaling
    A = src_geobox.transform * Affine.scale(scaler, scaler)

    return GeoBox((ny, nx), A, src_geobox.crs)


def _round_to_res(value: float, res: float) -> int:
    res = abs(res)
    return int(math.ceil((value - 0.1 * res) / res))


def flipy(gbox: GeoBox) -> GeoBox:
    """Alias for :py:meth:`odc.geo.geobox.flipy`."""
    return gbox.flipy()


def flipx(gbox: GeoBox) -> GeoBox:
    """Alias for :py:meth:`odc.geo.geobox.flipx`."""
    return gbox.flipx()


def translate_pix(gbox: GeoBox, tx: float, ty: float) -> GeoBox:
    """Alias for :py:meth:`odc.geo.geobox.GeoBox.translate_pix`."""
    return gbox.translate_pix(tx, ty)


def pad(gbox: GeoBox, padx: int, pady: MaybeInt = None) -> GeoBox:
    """Alias for :py:meth:`odc.geo.geobox.GeoBox.pad`."""
    return gbox.pad(padx, pady)


def pad_wh(gbox: GeoBox, alignx: int = 16, aligny: MaybeInt = None) -> GeoBox:
    """Alias for :py:meth:`odc.geo.geobox.GeoBox.pad_wh`."""
    return gbox.pad_wh(alignx, aligny)


def zoom_out(gbox: GeoBox, factor: float) -> GeoBox:
    """Alias for :py:meth:`odc.geo.geobox.GeoBox.zoom_out`."""
    return gbox.zoom_out(factor)


def zoom_to(
    gbox: GeoBox,
    shape: Union[SomeShape, int, float, None] = None,
    *,
    resolution: Optional[SomeResolution] = None,
) -> GeoBox:
    """Alias for :py:meth:`odc.geo.geobox.GeoBox.zoom_to`."""
    return gbox.zoom_to(shape, resolution=resolution)


def rotate(gbox: GeoBox, deg: float) -> GeoBox:
    """Alias for :py:meth:`odc.geo.geobox.GeoBox.`."""
    return gbox.rotate(deg)


def affine_transform_pix(gbox: GeoBox, transform: Affine) -> GeoBox:
    """Alias for :py:meth:`odc.geo.geobox.GeoBox.__mul__`."""
    return gbox * transform


class GeoboxTiles:
    """Partition GeoBox into sub geoboxes."""

    __slots__ = ("_gbox", "_tiles")

    def __init__(
        self,
        box: GeoBoxBase,
        tile_shape: Union[SomeShape, Chunks2d, None],
        *,
        _tiles: Optional[RoiTiles] = None,
    ):
        """
        Construct from a :py:class:`~odc.geo.GeoBox`.

        :param box: source :py:class:`~odc.geo.GeoBox`
        :param tile_shape: Shape of sub-tiles in pixels ``(rows, cols)``
        """
        self._gbox = box
        if _tiles is not None:
            self._tiles = _tiles
        else:
            assert tile_shape is not None
            self._tiles = roi_tiles(box.shape, tile_shape)

    @property
    def base(self) -> GeoBoxBase:
        """Access base Geobox"""
        return self._gbox

    @property
    def shape(self) -> Shape2d:
        """Number of tiles along each dimension."""
        return self._tiles.shape

    @property
    def roi(self) -> RoiTiles:
        """
        Access ROI covered by tile.

        .. code-block:: python

            gbt = GeoboxTiles(..)
            roi = gbt.roi[0, 3]
        """
        return self._tiles

    def chunk_shape(self, idx: SomeIndex2d) -> Shape2d:
        """
        Query chunk shape for a given chunk.

        :param idx: ``(row, col)`` chunk index
        :returns: ``(nrows, ncols)`` shape of a tile (edge tiles might be smaller)
        :raises: :py:class:`IndexError` when index is outside of ``[(0,0) -> .shape)``.
        """
        return self._tiles.tile_shape(idx)

    @property
    def chunks(self) -> Chunks2d:
        return self._tiles.chunks

    def _crop(self, roi: ROI) -> "GeoboxTiles":
        gbox_new = self.base[self._tiles[roi]]
        return GeoboxTiles(gbox_new, (0, 0), _tiles=self._tiles.crop(roi))

    def clip(
        self, selection: Sequence[Tuple[int, int]]
    ) -> Tuple["GeoboxTiles", List[Tuple[int, int]]]:
        """
        Crop to a set of tiles.

        Returns cropped version of :py:class:`GeoboxTiles` and a list of
        tile coordinates in the new, cropped space.
        """
        tiles, roi, new_idx = clip_tiles(self._tiles, selection)
        return GeoboxTiles(self[roi], None, _tiles=tiles), new_idx

    @property
    def crop(self) -> Mapping[ROI, "GeoboxTiles"]:
        return func2map(self._crop)

    def __getitem__(self, idx: Union[SomeIndex2d, ROI]) -> GeoBoxBase:
        """
        Lookup tile by index, index is in matrix access order: ``(row, col)``.

        :param idx: ``(row, col)`` index
        :returns: GeoBox of a tile
        :raises: IndexError when index is outside of ``[(0,0) -> .shape)``
        """
        return self._gbox[self._tiles[idx]]

    def pix_bbox(self, idx: Union[SomeIndex2d, ROI]) -> BoundingBox:
        """
        BoundingBox in pixel space of a given tile.
        """
        ry, rx = self._tiles[idx]
        return BoundingBox(rx.start, ry.start, rx.stop, ry.stop)

    def range_from_bbox(self, bbox: BoundingBox) -> Tuple[range, range]:
        """
        Intersect with a bounding box.

        Compute rows and columns overlapping with a given :py:class:`~odc.geo.geom.BoundingBox`.
        """

        if bbox.crs is not None:
            bbox = self._gbox.project(bbox.polygon).boundingbox

        def _clamp(span: Tuple[float, float], N: int):
            a1, a2 = span
            a1 = int(clamp(math.floor(a1), 0, N - 1))
            a2 = int(clamp(math.ceil(a2), 1, N)) - 1
            return a1, a2

        NY, NX = self._gbox.shape.yx
        x1, x2 = _clamp(bbox.range_x, NX)
        y1, y2 = _clamp(bbox.range_y, NY)

        y1, x1 = self._tiles.locate((y1, x1))
        y2, x2 = self._tiles.locate((y2, x2))

        return range(y1, y2 + 1), range(x1, x2 + 1)

    def _tiles_from_pix_bbox(self, bbox: BoundingBox) -> Iterator[Tuple[int, int]]:
        yy, xx = self.range_from_bbox(bbox)
        yield from itertools.product(yy, xx)

    def tiles(self, query: Union[Geometry, BoundingBox]) -> Iterator[Tuple[int, int]]:
        """Return tile indexes overlapping with a given geometry."""
        target_crs = self._gbox.crs

        if isinstance(query, BoundingBox):
            if query.crs is None:
                # special case for bounding box in pixel domain
                yield from self._tiles_from_pix_bbox(query)
                return
            poly = query.polygon
        else:
            poly = query

        if target_crs is not None and poly.crs != target_crs:
            poly = poly.to_crs(target_crs, check_and_fix=True)

        yy, xx = self.range_from_bbox(poly.boundingbox)
        for idx in itertools.product(yy, xx):
            gbox = self[idx]
            if not poly.disjoint(gbox.extent):
                yield idx

    def _check_linear(self, src: "GeoboxTiles") -> Optional[Affine]:
        if src.base.crs != self.base.crs:
            return None
        if not isinstance(self.base, GeoBox):
            return None
        if not isinstance(src.base, GeoBox):
            return None
        # src_pix = A*dst_pix
        A = snap_affine((~src.base.transform) * self.base.transform)
        if is_affine_st(A):
            return A
        return None

    def _all_tiles(self) -> Iterator[Tuple[int, int]]:
        for iy, ix in numpy.ndindex(self.shape.shape):
            yield (iy, ix)

    def _grid_intersect_linear(
        self,
        src: "GeoboxTiles",
        A: Affine,
    ) -> Dict[Tuple[int, int], List[Tuple[int, int]]]:
        deps: Dict[Tuple[int, int], List[Tuple[int, int]]] = {}

        for idx in self._all_tiles():
            bbox = self.pix_bbox(idx).transform(A).round()
            src_idx = list(src.tiles(bbox))
            deps[idx] = src_idx

        return deps

    def grid_intersect(
        self, src: "GeoboxTiles"
    ) -> Dict[Tuple[int, int], List[Tuple[int, int]]]:
        """
        Figure out tile to tile overlap graph between two grids.

        For every tile in this :py:class:`GeoboxTiles` find every tile in ``other`` that
        intersects with this ``tile``.
        """
        A = self._check_linear(src)
        if A is not None:
            return self._grid_intersect_linear(src, A)

        if src.base.crs == self.base.crs:
            src_footprint = src.base.extent
        else:
            # compute "robust" source footprint in CRS of self via espg:4326
            src_footprint = (
                src.base.footprint(4326, 2) & self.base.footprint(4326, 2)
            ).to_crs(self.base.crs)

        xy_chunks_with_data = list(self.tiles(src_footprint))
        deps: Dict[Tuple[int, int], List[Tuple[int, int]]] = {}

        for idx in xy_chunks_with_data:
            geobox = self[idx]
            deps[idx] = list(src.tiles(geobox.extent))

        return deps

    def __dask_tokenize__(self):
        return (
            "odc.geo.geobox.GeoboxTiles",
            *self._gbox.__dask_tokenize__()[1:],
            *self._tiles.__dask_tokenize__()[1:],
        )

    def __str__(self):
        return str(self.roi)

    def __eq__(self, __value: object) -> bool:
        if not isinstance(__value, GeoboxTiles):
            return False
        if self is __value:
            return True
        return self._tiles == __value._tiles and self._gbox == __value._gbox

    __repr__ = __str__



import pytest
from affine import Affine
from pyproj.aoi import AreaOfInterest

from odc.geo import BoundingBox, wh_
from odc.geo.geom import bbox_intersection, bbox_union
from odc.geo.testutils import epsg3857


def test_boundingbox():
    bb = BoundingBox(0, 3, 2, 4)
    assert bb.width == 2
    assert bb.height == 1
    assert bb.width == bb.span_x
    assert bb.height == bb.span_y
    assert bb.shape == (bb.height, bb.width)
    assert bb.crs is None
    assert bb.bbox == (0, 3, 2, 4)
    assert "crs=" not in str(bb)
    assert "crs=" not in repr(bb)
    assert bb.aoi == AreaOfInterest(0, 3, 2, 4)

    bb = BoundingBox(0, 3, 2.1, 4, "epsg:4326")
    assert bb.width == 2
    assert bb.height == 1
    assert bb.span_x == 2.1
    assert bb.width != bb.span_x
    assert bb.height == bb.span_y
    assert bb.crs.epsg == 4326
    assert "crs=" in str(bb)
    assert "crs=" in repr(bb)
    assert bb.aoi == AreaOfInterest(0, 3, 2.1, 4)

    assert BoundingBox.from_xy(bb.range_x, bb.range_y, bb.crs) == bb

    assert BoundingBox.from_xy((1, 2), (10, 20)) == (1, 10, 2, 20)
    assert BoundingBox.from_xy((2, 1), (20, 10)) == (1, 10, 2, 20)
    assert BoundingBox.from_points((1, 11), (2, 22)) == (1, 11, 2, 22)
    assert BoundingBox.from_points((1, 22), (2, 11)) == (1, 11, 2, 22)

    assert BoundingBox(1, 3, 10, 20).buffered(1, 3) == (0, 0, 11, 23)
    assert BoundingBox(1, 3, 10, 20).buffered(1) == (0, 2, 11, 21)

    assert BoundingBox.from_transform(
        wh_(10, 12), Affine.translation(1, 3), "epsg:3857"
    ) == (1, 3, 11, 15)
    assert (
        BoundingBox.from_transform(
            wh_(10, 12), Affine.translation(1, 3), "epsg:3857"
        ).crs
        == "epsg:3857"
    )

    assert bb.polygon.boundingbox == bb

    bb = BoundingBox(0, 3, 2, 4, "epsg:3857")
    assert bb.aoi == bb.to_crs("epsg:4326").aoi
    assert bb.aoi != AreaOfInterest(0, 3, 2, 4)


@pytest.mark.parametrize("crs", [None, "epsg:4326", epsg3857])
def test_bbox_union(crs):
    b1 = BoundingBox(0, 1, 10, 20, crs)
    b2 = BoundingBox(5, 6, 11, 22, crs)

    assert bbox_union([b1]) == b1
    assert bbox_union([b2]) == b2

    bb = bbox_union(iter([b1, b2]))
    assert bb == BoundingBox(0, 1, 11, 22, crs)

    bb = bbox_union(iter([b2, b1] * 10))
    assert bb == BoundingBox(0, 1, 11, 22, crs)

    assert b2 | b1 == bbox_union([b2, b1])
    assert b1 | b2 == bbox_union([b2, b1])


@pytest.mark.parametrize("crs", [None, "epsg:4326", epsg3857])
def test_bbox_intersection(crs):
    b1 = BoundingBox(0, 1, 10, 20, crs)
    b2 = BoundingBox(3, -10, 12, 8, crs)

    assert b1 & b2 == bbox_intersection([b1, b2])
    assert b2 & b1 == bbox_intersection([b1, b2])
    assert b1 & b1 == b1
    assert b2 & b2 == b2
    assert bbox_intersection([b1, b2]) == BoundingBox(3, 1, 10, 8, crs)
    assert b1 & b2 == BoundingBox(3, 1, 10, 8, crs)


@pytest.mark.parametrize(
    "crss",
    [
        (None, "epsg:4326", epsg3857),
        ("epsg:4326", "epsg:3857"),
        (*["epsg:4326"] * 4, "epsg:3857"),
    ],
)
def test_bbox_crs_mismatch(crss):
    with pytest.raises(ValueError):
        _ = bbox_union(BoundingBox(0, 0, 1, 1, crs) for crs in crss)

    with pytest.raises(ValueError):
        _ = bbox_intersection(BoundingBox(0, 0, 1, 1, crs) for crs in crss)


def test_map_bounds():
    bbox = BoundingBox(-180, -90, 180, 90, "epsg:4326")
    assert bbox.map_bounds() == ((-90, -180), (90, 180))

    bbox = BoundingBox(-180, -90, 180, 90)
    assert bbox.map_bounds() == ((-90, -180), (90, 180))

    bbox = BoundingBox(-10, -20, 24, 0, "epsg:4326")
    assert bbox.to_crs("epsg:3857").map_bounds() == ((-20, -10), (0, 24))


def test_bbox_to_crs():
    bbox = BoundingBox(-10, -20, 100, 0, "epsg:4326")
    assert bbox.to_crs("epsg:3857") == bbox.polygon.to_crs("epsg:3857").boundingbox


@pytest.mark.parametrize(
    "bb,expect",
    [
        ((-10, -20, 100, 0), (-10, -20, 100, 0)),
        ((-10.1, 2.3, 10.9, 2.4), (-11, 2, 11, 3)),
        ((-10.9, 2.6, 10.9, 2.9), (-11, 2, 11, 3)),
        ((0.9, 1.9, 1.01, 5.02), (0, 1, 2, 6)),
    ],
)
@pytest.mark.parametrize("crs", ["epsg:4326", None, "epsg:3857"])
def test_round(bb, expect, crs):
    expect = BoundingBox(*expect, crs)
    bbox = BoundingBox(*bb, crs)

    assert bbox.round().round() == bbox.round()
    assert bbox.round().polygon.contains(bbox.polygon)
    assert expect == bbox.round()



# pylint: disable=wrong-import-position,redefined-outer-name
from pathlib import Path
from unittest.mock import MagicMock
from warnings import catch_warnings, filterwarnings

import pytest

rasterio = pytest.importorskip("rasterio")
gpd = pytest.importorskip("geopandas")

from odc.geo._interop import have
from odc.geo.converters import extract_gcps, from_geopandas, map_crs, rio_geobox
from odc.geo.gcp import GCPGeoBox
from odc.geo.geobox import GeoBox, GeoBoxBase


@pytest.fixture
def countries_geodataframe():
    from odc.geo.data import Countries

    yield Countries()._instance


def test_from_geopandas(countries_geodataframe):
    df = countries_geodataframe
    gg = from_geopandas(df)
    assert isinstance(gg, list)
    assert len(gg) == len(df)
    assert gg[0].crs == "epsg:4326"

    (au,) = from_geopandas(df[df.ISO_A3 == "AUS"].to_crs(epsg=3577))
    assert au.crs.epsg == 3577

    (au,) = from_geopandas(df[df.ISO_A3 == "AUS"].to_crs(epsg=3857).geometry)
    assert au.crs.epsg == 3857

    assert from_geopandas(df.CONTINENT) == []


def test_have():
    assert isinstance(have.geopandas, bool)
    assert isinstance(have.rasterio, bool)
    assert isinstance(have.xarray, bool)
    assert isinstance(have.dask, bool)
    assert isinstance(have.folium, bool)
    assert isinstance(have.ipyleaflet, bool)
    assert isinstance(have.datacube, bool)
    assert isinstance(have.tifffile, bool)
    assert have.rasterio is True
    assert have.xarray is True
    assert have.check_or_error("xarray", "rasterio") is None

    with pytest.raises(RuntimeError):
        have.check_or_error("xarray", "noSuchLibaBCD")


def test_extract_gcps(data_dir: Path):
    with rasterio.open(data_dir / "au-gcp.tif") as src:
        assert src.gcps != ([], None)
        pix1, wld_native = extract_gcps(src)
        pix2, wld_3857 = extract_gcps(src, "epsg:3857")

    assert pix1 == pix2
    assert all(pt.crs == "epsg:4326" for pt in wld_native)
    assert all(pt.crs == "epsg:3857" for pt in wld_3857)

    with rasterio.open(data_dir / "au-3577.tif") as src:
        assert src.gcps == ([], None)
        with pytest.raises(ValueError):
            _ = extract_gcps(src)


@pytest.mark.parametrize("fname", ["au-gcp.tif", "au-3577.tif", "au-3577-rotated.tif"])
def test_rio_geobox(data_dir: Path, fname: str):
    with rasterio.open(data_dir / fname) as rdr:
        pts, _ = rdr.gcps
        gbox = rio_geobox(rdr)
        assert isinstance(gbox, GeoBoxBase)
        assert gbox.width == rdr.width
        assert gbox.height == rdr.height

        if len(pts) > 0:
            assert isinstance(gbox, GCPGeoBox)
        else:
            assert isinstance(gbox, GeoBox)


def test_map_crs():
    proj_3031 = "+proj=stere +lat_0=-90 +lat_ts=-71 +lon_0=0 +x_0=0 +y_0=0 +datum=WGS84 +units=m +no_defs +type=crs"
    assert map_crs(MagicMock(crs="EPSG4326")).epsg == 4326
    assert map_crs(MagicMock(crs=dict(name="EPSG3857"))).epsg == 3857
    assert map_crs(MagicMock(crs=dict(name="custom", proj4def=proj_3031))).epsg == 3031



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
from typing import Callable, Tuple, Union

import numpy as np
import xarray as xr
from affine import Affine

from . import CRS
from .geobox import GeoBox
from .gridspec import GridSpec
from .math import apply_affine, approx_equal_affine

# pylint: disable=invalid-name,

epsg4326 = CRS("EPSG:4326")
epsg3577 = CRS("EPSG:3577")
epsg3857 = CRS("EPSG:3857")
esri54019 = CRS("ESRI:54019")
modis_crs = CRS(
    "+proj=sinu +lon_0=0 +x_0=0 +y_0=0 +ellps=WGS84 +datum=WGS84 +units=m +no_defs"
)

AlbersGS = GridSpec(
    crs=epsg3577,
    tile_shape=(4000, 4000),
    resolution=25,
)

SAMPLE_WKT_WITHOUT_AUTHORITY = """PROJCS["unnamed",
       GEOGCS["unnamed ellipse",
              DATUM["unknown",
                    SPHEROID["unnamed",6378137,0],
                    EXTENSION["PROJ4_GRIDS","@null"]],
              PRIMEM["Greenwich",0],
              UNIT["degree",0.0174532925199433]],
       PROJECTION["Mercator_2SP"],
       PARAMETER["standard_parallel_1",0],
       PARAMETER["central_meridian",0],
       PARAMETER["false_easting",0],
       PARAMETER["false_northing",0],
       UNIT["Meter",1]
]
"""


def mkA(rot=0, scale=(1, 1), shear=0, translation=(0, 0)):
    return (
        Affine.translation(*translation)
        * Affine.rotation(rot)
        * Affine.shear(shear)
        * Affine.scale(*scale)
    )


def xy_from_gbox(gbox: GeoBox) -> Tuple[np.ndarray, np.ndarray]:
    """
    :returns: Two images with X and Y coordinates for centers of pixels
    """
    h, w = gbox.shape

    xx, yy = np.meshgrid(
        np.arange(w, dtype="float64") + 0.5, np.arange(h, dtype="float64") + 0.5
    )

    return apply_affine(gbox.transform, xx, yy)


def xy_norm(
    x: np.ndarray, y: np.ndarray, deg: float = 33.0
) -> Tuple[np.ndarray, np.ndarray, Affine]:
    """
    Transform output of xy_from_geobox with a reversible linear transform. On
    output x,y are in [0,1] range. Reversible Affine transform includes
    rotation by default, this is to ensure that test images don't have
    symmetries that are aligned to X/Y axis.

    1. Rotate x,y by ``deg``
    2. Offset and scale such that values are in [0, 1] range


    :returns: (x', y', A)

    - (x, y) == A*(x', y')
    - [x|y]'.min() == 0
    - [x|y]'.max() == 1

    """

    def norm_v(v):
        vmin = v.min()
        v -= vmin
        s = 1.0 / v.max()
        v *= s

        return (s, -vmin * s)

    A_rot = Affine.rotation(deg)
    x, y = apply_affine(A_rot, x, y)

    sx, tx = norm_v(x)
    sy, ty = norm_v(y)

    A = Affine(sx, 0, tx, 0, sy, ty) * A_rot

    return x, y, ~A


def to_fixed_point(a, dtype="uint16"):
    """
    Convert normalised ([0,1]) floating point image to integer fixed point fractional.

    Note for signed types: there is no offset, 0 -> 0, 1 -> (2**(nbits - 1) - 1).

    Reverse is provided by: ``from_fixed_point``
    """
    ii = np.iinfo(dtype)
    a = a * ii.max + 0.5
    a = np.clip(a, 0, ii.max, out=a)
    return a.astype(ii.dtype)


def from_fixed_point(a):
    """
    Convert fixed point image to floating point

    This is reverse of ``to_fixed_point``
    """
    ii = np.iinfo(a.dtype)
    return a.astype("float64") * (1.0 / ii.max)


def gen_test_image_xy(
    gbox: GeoBox, dtype: Union[str, np.dtype, type] = "float32", deg: float = 33.0
) -> Tuple[np.ndarray, Callable]:
    """
    Generate test image that captures pixel coordinates in pixel values.
    Useful for testing reprojections/reads.

    :param gbox: GeoBox defining pixel plane

    :dtype: data type of the image, defaults to `float32`, but it can be an
            integer type in which case normalised coordinates will be
            quantised increasing error.

    :returns: 2xWxH ndarray encoding X,Y coordinates of pixel centers in some
              normalised space, and a callable that can convert from normalised
              space back to coordinate space.

    """
    dtype = np.dtype(dtype)

    x, y = xy_from_gbox(gbox)
    x, y, A = xy_norm(x, y, deg)

    xy = np.stack([x, y])

    if dtype.kind == "f":
        xy = xy.astype(dtype)
    else:
        xy = to_fixed_point(xy, dtype)

    def denorm(xy=None, y=None, nodata=None):
        if xy is None:
            return A

        stacked = y is None
        x, y = xy if stacked else (xy, y)
        missing_mask = None

        if nodata is not None:
            if np.isnan(nodata):
                missing_mask = np.isnan(x) + np.isnan(y)
            else:
                missing_mask = (x == nodata) + (y == nodata)

        if x.dtype.kind != "f":
            x = from_fixed_point(x)
            y = from_fixed_point(y)

        x, y = apply_affine(A, x, y)

        if missing_mask is not None:
            x[missing_mask] = np.nan
            y[missing_mask] = np.nan

        if stacked:
            return np.stack([x, y])

        return x, y

    return xy, denorm


def purge_crs_info(xx: xr.DataArray) -> xr.DataArray:
    attributes_to_clear = ["crs", "wkt", "crs_wkt", "transform", "epsg", "resolution"]

    def _purge_attributes(_x: xr.DataArray) -> xr.DataArray:
        for attr in attributes_to_clear:
            _x.attrs.pop(attr, None)
        _x.encoding.pop("grid_mapping", None)
        return _x

    # remove non-dimensional coordinate, which is CRS in our case
    to_drop = [name for name in xx.coords if name not in xx.dims]
    xx = xx.drop_vars(to_drop)
    xx.encoding.pop("grid_spatial", None)

    # purge attributes from coords and xx
    cc = {name: _purge_attributes(coord) for name, coord in xx.coords.items()}
    return _purge_attributes(xx.assign_coords(cc))


def daskify(xx, **kw):
    # pylint: disable=import-outside-toplevel
    from dask import array as da

    if isinstance(xx, xr.Dataset):
        return xr.Dataset({name: daskify(v) for name, v in xx.data_vars.items()})

    return xr.DataArray(
        data=da.from_array(xx.data, **kw),
        dims=xx.dims,
        coords=xx.coords,
        attrs=xx.attrs,
    )


def approx_equal_geobox(a: GeoBox, b: GeoBox, tol: float = 1e-6) -> bool:
    if a.shape != b.shape or a.crs != b.crs:
        return False

    return approx_equal_affine(a.transform, b.transform, tol)



"""Tests for odc.geo.cog._s3."""

import unittest

from odc.geo.cog._s3 import S3MultiPartUpload

# Conditional import for S3 support
try:
    from odc.geo.cog._s3 import S3MultiPartUpload

    HAVE_S3 = True
except ImportError:
    S3MultiPartUpload = None
    HAVE_S3 = False


def require_s3(test_func):
    """Decorator to skip tests if s3 dependencies are not installed."""
    return unittest.skipUnless(HAVE_S3, "s3 dependencies are not installed")(test_func)


@require_s3
def test_s3_mpu():
    """Test S3MultiPartUpload class initialization."""
    mpu = S3MultiPartUpload("bucket", "file.dat")
    if mpu.bucket != "bucket":
        raise ValueError("Invalid bucket")
    if mpu.key != "file.dat":
        raise ValueError("Invalid key")



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""GridSpec class."""
import math
from typing import Any, Dict, Iterator, Optional, Tuple

from affine import Affine

from . import geom
from .crs import SomeCRS, norm_crs_or_error
from .geobox import GeoBox
from .geom import BoundingBox, Geometry
from .math import Bin1D
from .types import (
    XY,
    Index2d,
    Shape2d,
    SomeIndex2d,
    SomeResolution,
    SomeShape,
    ixy_,
    res_,
    resyx_,
    shape_,
    xy_,
    yx_,
)


class GridSpec:
    """
    Definition for a regular spatial grid.

    :param crs:
       Coordinate System used to define the grid
    :param tile_shape:
       Size of each tile in pixels
    :param resolution:
       Size of each data point in the grid, in CRS units. ``Y`` will usually be negative.
    :param origin:
       Coordinate of a bottom-left corner of the ``(0,0)`` tile in CRS units. Default is
       ``xy_(0.0, 0.0)``
    :param flipx: when ``True`` grid index for X axis increments left to right
    :param flipy: when ``True`` grid index for Y axis increments top to bottom
    """

    def __init__(
        self,
        crs: SomeCRS,
        tile_shape: SomeShape,
        resolution: SomeResolution,
        origin: Optional[XY[float]] = None,
        flipx: bool = False,
        flipy: bool = False,
    ):
        tile_shape = shape_(tile_shape)
        resolution = res_(resolution)

        if origin is None:
            origin = xy_(0.0, 0.0)
        else:
            assert isinstance(origin, XY)

        self.crs = norm_crs_or_error(crs)
        self._shape = tile_shape
        self.resolution = resolution
        self.tile_size = xy_(
            tile_shape.x * abs(resolution.x),
            tile_shape.y * abs(resolution.y),
        )
        self.origin = origin

        ox, oy = origin.xy
        self._ybin = Bin1D(self.tile_size.y, oy, -1 if flipy else 1)
        self._xbin = Bin1D(self.tile_size.x, ox, -1 if flipx else 1)

    def __eq__(self, other):
        if not isinstance(other, GridSpec):
            return False

        return (
            self._shape == other._shape
            and self._ybin == other._ybin
            and self._xbin == other._xbin
            and self.crs == other.crs
        )

    @property
    def dimensions(self) -> Tuple[str, str]:
        """List of dimension names of the grid spec."""
        return self.crs.dimensions

    @property
    def alignment(self) -> XY[float]:
        """Pixel boundary alignment."""
        y, x = (
            orig % abs(res) for orig, res in zip(self.origin.yx, self.resolution.yx)
        )
        return yx_(y, x)

    @property
    def tile_shape(self) -> Shape2d:
        """Tile shape in pixels (Y,X order, like numpy)."""
        return self._shape

    def pt2idx(self, x: float, y: float) -> Index2d:
        """
        Compute tile index from a point.

        :param x: X coordinate of a point in CRS units
        :param y: Y coordinate of a point in CRS units
        :return:
          ``(ix, iy)``, index of a tile containing given point
        """
        return ixy_(self._xbin.bin(x), self._ybin.bin(y))

    def _tile_txy(self, tile_index: Index2d) -> XY[float]:
        """Location of 0,0 pixel in CRS units."""

        ix, iy = tile_index.xy
        rx, ry = self.resolution.xy

        x0, x1 = self._xbin[ix]
        tx = x0 if rx > 0 else x1

        y0, y1 = self._ybin[iy]
        ty = y0 if ry > 0 else y1

        return xy_(tx, ty)

    def tile_geobox(self, tile_index: SomeIndex2d) -> GeoBox:
        """
        Tile geobox.

        :param tile_index: ``(ix, iy)``
        """
        tile_index = ixy_(tile_index)
        tx, ty = self._tile_txy(tile_index).xy
        rx, ry = self.resolution.xy
        return GeoBox(self._shape, crs=self.crs, affine=Affine(rx, 0, tx, 0, ry, ty))

    def __getitem__(self, idx: SomeIndex2d) -> GeoBox:
        """Lookup :py:class:`~odc.geo.geobox.GeoBox` of a given tile."""
        return self.tile_geobox(idx)

    def idx_bounds(self, bounds: BoundingBox) -> Tuple[int, int, int, int]:
        """
        Convert bounds from CRS to index space.

        :param bounds: Query bounding box
        :return: ``x1, y1, x2, y2``, with closed/open range, i.e. ``[x1, x2), [y1, y2)``
        """
        assert self.crs == bounds.crs
        tol = 1e-8

        x1, y1, x2, y2 = bounds
        ix1, iy1 = self.pt2idx(x1 + tol, y1 + tol).xy
        ix2, iy2 = self.pt2idx(x2 - tol, y2 - tol).xy

        ix1, ix2 = sorted([ix1, ix2])
        iy1, iy2 = sorted([iy1, iy2])
        return ix1, iy1, ix2 + 1, iy2 + 1

    def tiles(
        self, bounds: BoundingBox, geobox_cache: Optional[dict] = None
    ) -> Iterator[Tuple[Tuple[int, int], GeoBox]]:
        """
        Query tiles overlapping with bounding box.

        Output is a sequence of ``tile_index``, :py:class:`odc.geo.geobox.GeoBox` tuples.

        .. note::

           Grid cells are referenced by coordinates ``(x, y)``, which is the opposite to the usual
           CRS dimension order.

        :param bounds:
           Boundary coordinates of the required grid
        :param geobox_cache:
           Optional cache to re-use geoboxes instead of creating new one each time
        :return:
          Iterator of tuples of grid index and corresponding :py:class:`odc.geo.geobox.GeoBox`
        """

        def geobox(tile_index):
            if geobox_cache is None:
                return self.tile_geobox(tile_index)

            gbox = geobox_cache.get(tile_index)
            if gbox is None:
                gbox = self.tile_geobox(tile_index)
                geobox_cache[tile_index] = gbox
            return gbox

        ix1, iy1, ix2, iy2 = map(int, self.idx_bounds(bounds))

        for iy in range(iy1, iy2):
            for ix in range(ix1, ix2):
                tile_index = (ix, iy)
                yield tile_index, geobox(tile_index)

    def tiles_from_geopolygon(
        self,
        geopolygon: Geometry,
        geobox_cache: Optional[dict] = None,
    ) -> Iterator[Tuple[Tuple[int, int], GeoBox]]:
        """
        Query tiles overlapping with a given polygon.

        Output is a sequence of ``tile_index``, :py:class:`odc.geo.geobox.GeoBox` tuples.

        :param geopolygon:
          Polygon to tile
        :param geobox_cache:
          Optional cache to re-use geoboxes instead of creating new one each turn: iterator of grid
          cells with :py:class:`odc.geo.geobox.GeoBox` tiles
        """
        geopolygon = geopolygon.to_crs(self.crs, check_and_fix=True)
        bbox = geopolygon.boundingbox

        for tile_index, tile_geobox in self.tiles(bbox, geobox_cache):
            if not geopolygon.disjoint(tile_geobox.extent):
                yield (tile_index, tile_geobox)

    def __str__(self) -> str:
        return f"GridSpec(crs={self.crs}, tile_shape={self._shape}, resolution={self.resolution})"

    def __repr__(self) -> str:
        return self.__str__()

    def geojson(
        self,
        *,
        bbox: Optional[BoundingBox] = None,
        geopolygon: Optional[Geometry] = None,
    ) -> Dict[str, Any]:
        """
        Render to GeoJSON.

        :param bbox:
           Limit output to tiles overlapping with the given bounding box (native CRS of the grid).

        :param geopolygon:
           Limit output to tiles overlapping with the given geometry (any CRS)

        :return: GeoJSON representation of the grid spec
        """
        if geopolygon is not None:
            _tiles = self.tiles_from_geopolygon(geopolygon)
        elif bbox is not None:
            _tiles = self.tiles(bbox)
        else:
            valid_region = self.crs.valid_region
            if valid_region is None:
                valid_region = geom.box(-180, -90, 180, 90, "epsg:4326")

            _tiles = self.tiles(
                valid_region.buffer(-0.05).to_crs(self.crs, resolution=0.5).boundingbox
            )

        props = {
            "native_crs": str(self.crs),
            "tile_shape": self._shape,
            "resolution": self.resolution,
        }

        features = []
        for (ix, iy), geobox in _tiles:
            features.append(geobox.extent.geojson(idx=f"{ix},{iy}"))
        return {"type": "FeatureCollection", "features": features, "properties": props}

    @staticmethod
    def from_sample_tile(
        box: Geometry,
        *,
        shape: SomeShape = (-1, -1),
        idx: SomeIndex2d = (0, 0),
        flipx: bool = False,
        flipy: bool = False,
    ) -> "GridSpec":
        """
        Construct :py:class:`odc.geo.gridspec.GridSpec` from a sample tile.

        Bounding box of one tile, it's index, and a shape of the tile in pixels fully define a grid.
        This method could be more convenient than canonical representation.

        :param box: Geometry of the tile in some CRS
        :param idx: ``ix, iy`` index of the sample tile, default ``(0, 0)``
        :param shape: ``height, width`` of the tile, must be supplied
        :param flipx: when ``True`` grid index for X axis increments left to right
        :param flipy: when ``True`` grid index for Y axis increments top to bottom
        """
        if shape == (-1, -1):
            raise ValueError("Must specify shape of the tile in pixels")
        shape = shape_(shape)
        idx = ixy_(idx)

        crs = norm_crs_or_error(box.crs)
        ix, iy = idx.xy
        nx, ny = shape.xy
        bbox = box.boundingbox

        xbin = Bin1D.from_sample_bin(ix, bbox.range_x, -1 if flipx else 1)
        ybin = Bin1D.from_sample_bin(iy, bbox.range_y, -1 if flipy else 1)

        origin = xy_(xbin.origin, ybin.origin)
        resolution = resyx_(-ybin.sz / ny, xbin.sz / nx)
        return GridSpec(
            crs, shape, resolution=resolution, origin=origin, flipx=flipx, flipy=flipy
        )

    @staticmethod
    def web_tiles(zoom: int, npix: int = 256) -> "GridSpec":
        """
        Construct :py:class:`~odc.geo.gridspec.GridSpec` that matches slippy tiles.

        Tile with index ``(0, 0)`` is at the top left corner of the map, and tile with index
        ``(2^zoom - 1, 2^zoom - 1)`` is at the bottom right.

        :param zoom:
           Zoom level, ``0`` is one single tile, ``1`` is ``2x2``, ``3`` is ``8x8``...

        :param npix:
           Usually tiles are ``256x256`` pixels wide. But you can override that.

        :return: Grid spec that encodes slippy tiles scheme in ``EPSG:3857``.
        """
        R = 6_378_137
        pi = math.pi
        tsz = pi * R * (2 ** (1 - zoom))  # in meters
        x, y = -pi * R, pi * R  # top-left corner of tile 0,0
        tile0 = geom.box(x, y - tsz, x + tsz, y, "epsg:3857")
        shape = (npix, npix)

        return GridSpec.from_sample_tile(tile0, shape=shape, idx=(0, 0), flipy=True)



# pylint: disable=wrong-import-position
import pytest

pytest.importorskip("dask")

import numpy as np
from dask.base import tokenize

from odc.geo.gcp import GCPGeoBox, GCPMapping
from odc.geo.geobox import CRS, GeoBox, GeoboxTiles


def test_tokenize_geobox():
    gbox = GeoBox.from_bbox([0, 0, 1, 1], shape=(100, 100))
    tk = tokenize(gbox)
    print(tk)
    assert tokenize(gbox) == tokenize(gbox)
    assert tokenize(gbox) != tokenize(gbox.pad(1))

    gbt = GeoboxTiles(gbox, (1, 3))
    assert tokenize(gbt) == tokenize(gbt)
    assert tokenize(GeoboxTiles(gbox, (1, 2))) != tokenize(GeoboxTiles(gbox, (2, 1)))
    assert tokenize(GeoboxTiles(gbox, (1, 2))) != tokenize(
        GeoboxTiles(gbox.pad(1), (1, 2))
    )

    gbt = GeoboxTiles(gbox, ((30, 40, 30), (45, 55)))
    assert tokenize(gbt) == tokenize(GeoboxTiles(gbox, gbt.chunks))
    assert tokenize(gbt) != tokenize(GeoboxTiles(gbox, gbt.chunks[::-1]))

    crs = CRS("epsg:4326")
    assert tokenize(crs) == tokenize(crs)
    assert tokenize(crs) == tokenize(CRS(crs))
    assert tokenize(CRS("epsg:4326")) == tokenize(CRS("EPSG:4326"))
    assert tokenize(CRS("epsg:4326")) != tokenize(CRS("EPSG:3857"))


def test_tokenize_gcpgeobox():
    pts = np.vstack([c.ravel() for c in np.meshgrid([0, 2, 3], [0, 1, 2, 5])]).T
    assert pts.shape[0] > 9
    assert pts.shape[1] == 2

    mapping = GCPMapping(pts, pts + 10, "epsg:3857")
    mapping2 = GCPMapping(pts, pts + 10, "epsg:4326")
    mapping_ = GCPMapping(mapping._pix, mapping._wld, mapping.crs)

    assert tokenize(mapping) == tokenize(mapping_)

    gbox = GCPGeoBox((10, 20), mapping)
    gbox_ = GCPGeoBox((10, 20), mapping_)

    gbox2 = GCPGeoBox((10, 20), mapping2)
    assert tokenize(gbox) == tokenize(gbox_)
    assert tokenize(gbox[:3, :2]) == tokenize(gbox[0:3, 0:2])

    assert tokenize(gbox) != tokenize(gbox2)



"""Basic types."""

from enum import Enum
from typing import (
    Callable,
    Generic,
    Iterable,
    Iterator,
    List,
    Literal,
    Mapping,
    Optional,
    Protocol,
    Sequence,
    Tuple,
    TypeVar,
    Union,
    cast,
    overload,
)

MaybeInt = Optional[int]
MaybeFloat = Optional[float]
FillValue = Union[float, int]
Nodata = Union[float, int, None]
SomeNodata = Union[float, int, str, None, Literal["auto"]]
T = TypeVar("T")
T1 = TypeVar("T1")
T2 = TypeVar("T2")
TK = TypeVar("TK")
TV = TypeVar("TV")


class Unset:
    """
    Marker for unset values.

    Used where ``None`` can be a valid value.
    """

    __slots__ = ()


class XY(Generic[T]):
    """
    Immutable container for anything X/Y.

    This class is used as a replacement for a plain tuple of two values that could be in X/Y or Y/X
    order.

    :param x: Value of type ``T`` for x
    :param y: Value of type ``T`` for y
    """

    __slots__ = ("_xy",)

    def __init__(self, x: T, y: T) -> None:
        self._xy = x, y

    def __eq__(self, other) -> bool:
        if not isinstance(other, XY):
            return False
        return self._xy == other._xy

    def __str__(self) -> str:
        return f"XY(x={self._xy[0]}, y={self._xy[1]})"

    def __repr__(self) -> str:
        return f"XY(x={self._xy[0]}, y={self._xy[1]})"

    def __hash__(self) -> int:
        return hash(self._xy)

    @property
    def x(self) -> T:
        """Access X value."""
        return self._xy[0]

    @property
    def y(self) -> T:
        """Access Y value."""
        return self._xy[1]

    @property
    def xy(self) -> Tuple[T, T]:
        """Convert to tuple in X,Y order."""
        return self._xy

    @property
    def yx(self) -> Tuple[T, T]:
        """Convert to tuple in Y,X order."""
        return self._xy[1], self._xy[0]

    @property
    def lon(self) -> T:
        """Access Longitude value (X)."""
        return self._xy[0]

    @property
    def lat(self) -> T:
        """Access Latitude value (Y)."""
        return self._xy[1]

    @property
    def lonlat(self) -> Tuple[T, T]:
        """Convert to tuple in Longitude,Latitude order."""
        return self._xy

    @property
    def latlon(self) -> Tuple[T, T]:
        """Convert to tuple in Latitude,Longitude order."""
        return self._xy[1], self._xy[0]

    @property
    def shape(self) -> Tuple[int, int]:
        """
        Interpret as ``shape`` (Y, X) order.

        Only valid for ``XY[int]`` case.

        :raises ValueError: when tuple contains anything but integer values
        :return: ``(y, x)``
        """
        x, y = self._xy
        if isinstance(x, int) and isinstance(y, int):
            return y, x
        raise ValueError("Expect (int, int) for shape")

    @property
    def wh(self) -> Tuple[int, int]:
        """
        Interpret as ``width, height``, (X, Y) order.

        Only valid for ``XY[int]`` case.

        :raises ValueError: when tuple contains anything but integer values
        :return: ``(x, y)``
        """
        x, y = self._xy
        if isinstance(x, int) and isinstance(y, int):
            return (x, y)
        raise ValueError("Expect (int, int) for wh")

    def map(self, op: Callable[[T], T2]) -> "XY[T2]":
        """
        Apply function to x and y and return new XY value.
        """
        return xy_(op(self.x), op(self.y))

    @property
    def aspect(self) -> float:
        """Aspect ratio (X/Y)."""
        return float(self.x) / float(self.y)  # type: ignore


class Resolution(XY[float]):
    """
    Resolution for X/Y dimensions.

    """

    def __init__(self, x: float, y: Optional[float] = None) -> None:
        if y is None:
            y = -x
        super().__init__(float(x), float(y))

    def __repr__(self) -> str:
        return f"Resolution(x={self.x:g}, y={self.y:g})"

    def __str__(self) -> str:
        return f"Resolution(x={self.x:g}, y={self.y:g})"


class Index2d(XY[int]):
    """
    2d index.
    """

    def __init__(self, x: int, y: int) -> None:
        super().__init__(x, y)

    def __repr__(self) -> str:
        return f"Index2d(x={self.x}, y={self.y})"

    def __str__(self) -> str:
        return f"Index2d(x={self.x}, y={self.y})"


class Shape2d(XY[int], Sequence[int]):
    """
    2d shape.

    Unlike other XY types, Shape2d does have canonical order: ``Y,X``.
    This class implements Mapping interfaces, so it can be used as input into
    ``numpy`` functions that accept shape parameter.
    It can also be compared directly to a tuple form.
    It can be concatenated with a tuple.
    """

    def __init__(self, x: int, y: int) -> None:
        super().__init__(x, y)

    def __repr__(self) -> str:
        return f"Shape2d(x={self.x}, y={self.y})"

    def __str__(self) -> str:
        return f"Shape2d(x={self.x}, y={self.y})"

    def __len__(self) -> int:
        return 2

    def __iter__(self) -> Iterator[int]:
        yield from self.shape

    def __getitem__(self, idx):
        return self.shape[idx]

    def __eq__(self, other) -> bool:
        if isinstance(other, tuple):
            return self.shape == other
        return super().__eq__(other)

    def __add__(self, other):
        return self.shape.__add__(other)

    def __radd__(self, other):
        return other + self.shape

    def shrink2(self) -> "Shape2d":
        return Shape2d(x=self._xy[0] // 2, y=self._xy[1] // 2)


SomeShape = Union[Tuple[int, int], XY[int], Shape2d, Index2d]
SomeIndex2d = Union[Tuple[int, int], XY[int], Index2d]
SomeResolution = Union[float, int, Resolution]
Chunks2d = Tuple[Tuple[int, ...], Tuple[int, ...]]


class SupportsCoords(Protocol[T]):
    """
    Needed for Point geometry -> XY conversion.
    """

    @property
    def coords(self) -> List[Tuple[T, T]]: ...


# fmt: off
@overload
def xy_(x: T, y: T, /) -> XY[T]: ...
@overload
def xy_(x: XY[T], y: Literal[None] = None, /) -> XY[T]: ...
@overload
def xy_(x: SupportsCoords[T], y: Literal[None] = None, /) -> XY[T]: ...
@overload
def xy_(x: Iterable[T], y: Literal[None] = None, /) -> XY[T]: ...
# fmt: on


def xy_(
    x: Union[T, XY[T], SupportsCoords[T], Iterable[T]],
    y: Optional[T] = None,
) -> XY[T]:
    """
    Construct from X,Y order.

    .. code-block:: python

       xy_(0, 1)
       xy_([0, 1])
       xy_(tuple([0, 1]))
       assert xy_(1, 3).x == 1
       assert xy_(1, 3).y == 3
    """
    if y is not None:
        return XY(x=cast(T, x), y=y)

    if isinstance(x, XY):
        return x

    if (coords := getattr(x, "coords", None)) is not None:
        ((x, y),) = cast(List[Tuple[T, T]], coords)
        return XY(x=x, y=y)

    if not isinstance(x, Iterable):
        raise ValueError("Expect 2 arguments or a single iterable.")

    x, y = cast(Iterable[T], x)
    return XY(x=x, y=y)


# fmt: off
@overload
def yx_(y: T, x: T, /) -> XY[T]: ...
@overload
def yx_(y: Iterable[T], x: Literal[None] = None, /) -> XY[T]: ...
@overload
def yx_(y: XY[T], x: Literal[None] = None, /) -> XY[T]: ...
# fmt: on


def yx_(y: Union[T, XY[T], Iterable[T]], x: Optional[T] = None, /) -> XY[T]:
    """
    Construct from Y,X order.

    .. code-block:: python

       yx_(0, 1)
       yx_([0, 1])
       yx_(tuple([0, 1]))
       assert yx_(1, 3).x == 3
       assert yx_(1, 3).y == 1
    """
    if x is not None:
        return XY(x=x, y=cast(T, y))

    if isinstance(y, XY):
        return y

    if not isinstance(y, Iterable):
        raise ValueError("Expect 2 arguments or a single iterable.")

    y, x = cast(Iterable[T], y)
    return XY(x=x, y=y)


def res_(x: Union[Resolution, float, int], /) -> Resolution:
    """Resolution for square pixels with inverted Y axis."""
    if isinstance(x, Resolution):
        return x
    if isinstance(x, (int, float)):
        return Resolution(float(x))
    raise ValueError(f"Unsupported input type: res_(x: {type(x)})")


def resxy_(x: float, y: float, /) -> Resolution:
    """Construct resolution from X,Y order."""
    return Resolution(x=x, y=y)


def resyx_(y: float, x: float, /) -> Resolution:
    """Construct resolution from Y,X order."""
    return Resolution(x=x, y=y)


# fmt: off
@overload
def ixy_(x: int, y: int, /) -> Index2d: ...
@overload
def ixy_(x: Tuple[int, int], y: Literal[None] = None, /) -> Index2d: ...
@overload
def ixy_(x: Index2d, y: Literal[None] = None, /) -> Index2d: ...
@overload
def ixy_(x: XY[int], y: Literal[None] = None, /) -> Index2d: ...
# fmt: on


def ixy_(
    x: Union[int, Tuple[int, int], XY[int], Index2d], y: Optional[int] = None, /
) -> Index2d:
    """Construct 2d index in X,Y order."""
    if y is not None:
        assert isinstance(x, int)
        return Index2d(x=x, y=y)
    if isinstance(x, tuple):
        x, y = x
        return Index2d(x=x, y=y)
    if isinstance(x, Index2d):
        return x
    if isinstance(x, XY):
        x, y = x.xy
        return Index2d(x=x, y=y)
    raise ValueError("Expect 2 values or a single tuple/XY/Index2d object")


# fmt: off
@overload
def iyx_(y: int, x: int, /) -> Index2d: ...
@overload
def iyx_(y: Tuple[int, int], x: Literal[None] = None, /) -> Index2d: ...
@overload
def iyx_(y: Index2d, x: Literal[None] = None, /) -> Index2d: ...
@overload
def iyx_(y: XY[int], x: Literal[None] = None, /) -> Index2d: ...
# fmt: on


def iyx_(
    y: Union[int, Tuple[int, int], XY[int], Index2d], x: Optional[int] = None, /
) -> Index2d:
    """Construct 2d index in Y,X order."""
    if x is not None:
        assert isinstance(y, int)
        return Index2d(x=x, y=y)
    if isinstance(y, tuple):
        y, x = y
        return Index2d(x=x, y=y)
    if isinstance(y, Index2d):
        return y
    if isinstance(y, XY):
        y, x = y.yx
        return Index2d(x=x, y=y)
    raise ValueError("Expect 2 values or a single tuple/XY/Index2d object")


def wh_(w: int, h: int, /) -> Shape2d:
    """Shape from width/height."""
    return Shape2d(x=w, y=h)


def shape_(x: Union[SomeShape, Tuple[int, ...]]) -> Shape2d:
    """Normalise shape representation."""
    if isinstance(x, Shape2d):
        return x
    if isinstance(x, XY):
        nx, ny = x.map(int).xy
        return Shape2d(x=nx, y=ny)
    if isinstance(x, Sequence):
        ny, nx = map(int, x)
        return Shape2d(x=nx, y=ny)
    raise ValueError(f"Input type not understood: {type(x)}")


# fmt: off
class NormalizedSlice(Protocol):
    """
    Type for ``slice`` with start/stop set to integer values.
    """
    @property
    def start(self) -> int: ...
    @property
    def stop(self) -> int: ...
    @property
    def step(self) -> Optional[int]: ...
# fmt: on

SomeSlice = Union[slice, int, NormalizedSlice]
"""
Slice index into ndarray or a single int.

Single index is equivalent to ``slice(idx, idx+1)``.
"""

NdROI = Union[SomeSlice, Tuple[SomeSlice, ...]]
"""
Any dimensional slice into ndarray.

This could be a single ``int`` or slice ``slice`` or a tuple of any number
of those things.
"""

ROI = Tuple[SomeSlice, SomeSlice]
"""2d slice into an image plane."""

NormalizedROI = Tuple[NormalizedSlice, NormalizedSlice]
"""Normalized 2d slice into an image plane."""

OutlineMode = Union[
    Literal["native"], Literal["pixel"], Literal["geo"], Literal["auto"]
]


class AnchorEnum(Enum):
    """
    Defines which way to snap geobox pixel grid.
    """

    EDGE = 0
    """Snap pixel edges to multiples of pixel size."""

    CENTER = 1
    """Snap pixel centers to multiples of pixel size."""

    FLOATING = 2
    """Turn off pixel snapping."""


class _Func2Map(Mapping[TK, TV]):
    """
    Turn ``f(K)`` into ``ff[K]``.
    """

    __slots__ = ("_func", "_keys")

    def __init__(self, func: Callable[[TK], TV], keys: Sequence[TK]):
        self._func = func
        self._keys = keys

    def __getitem__(self, idx: TK) -> TV:
        return self._func(idx)

    def __len__(self) -> int:
        return len(self._keys)

    def __iter__(self) -> Iterator[TK]:
        yield from self._keys


def func2map(
    f: Callable[[TK], TV],
    keys: Optional[Sequence[TK]] = None,
) -> Mapping[TK, TV]:
    """
    Turn ``f(K) -> V`` into ``ff[K] -> V``.

    :param f: Callable mapping ``K -> V``
    :param keys: Optional sequence of valid keys
    """
    return _Func2Map(f, [] if keys is None else keys)



"""
Multipart upload interface.

Defines the `MultiPartUploadBase` class for implementing multipart upload functionality.
This interface standardises methods for initiating, uploading, and finalising
multipart uploads across storage backends.
"""

from abc import ABC, abstractmethod
from typing import Any, Union, TYPE_CHECKING

from dask.delayed import Delayed
from ._mpu import get_mpu_kwargs, mpu_upload

if TYPE_CHECKING:
    # pylint: disable=import-outside-toplevel,import-error
    import dask.bag


class MultiPartUploadBase(ABC):
    """Abstract base class for multipart upload."""

    @abstractmethod
    def initiate(self, **kwargs) -> str:
        """Initiate a multipart upload and return an identifier."""

    @abstractmethod
    def write_part(self, part: int, data: bytes) -> dict[str, Any]:
        """Upload a single part."""

    @abstractmethod
    def finalise(self, parts: list[dict[str, Any]]) -> str:
        """Finalise the upload with a list of parts."""

    @abstractmethod
    def cancel(self, other: str = ""):
        """Cancel the multipart upload."""

    @property
    @abstractmethod
    def url(self) -> str:
        """Return the URL of the upload target."""

    @property
    @abstractmethod
    def started(self) -> bool:
        """Check if the multipart upload has been initiated."""

    @abstractmethod
    def writer(self, kw: dict[str, Any], *, client: Any = None) -> Any:
        """
        Return a Dask-compatible writer for multipart uploads.

        :param kw: Additional parameters for the writer.
        :param client: Dask client for distributed execution.
        """

    @abstractmethod
    def dask_name_prefix(self) -> str:
        """Return the dask name prefix specific to the backend."""

    def upload(
        self,
        chunks: Union["dask.bag.Bag", list["dask.bag.Bag"]],
        *,
        mk_header: Any = None,
        mk_footer: Any = None,
        user_kw: dict[str, Any] | None = None,
        writes_per_chunk: int = 1,
        spill_sz: int = 20 * (1 << 20),
        client: Any = None,
    ) -> Delayed:
        """High-level upload that calls mpu_upload under the hood."""
        kwargs = get_mpu_kwargs(
            mk_header=mk_header,
            mk_footer=mk_footer,
            user_kw=user_kw,
            writes_per_chunk=writes_per_chunk,
            spill_sz=spill_sz,
            client=client,
        )
        return mpu_upload(
            chunks,
            writer=self.writer,
            dask_name_prefix=self.dask_name_prefix(),
            **kwargs,
        )



""" Helpers for dealing with RGB(A) images.
"""

import functools
from typing import Any, List, Optional, Tuple

import numpy as np
import xarray as xr

from ._interop import is_dask_collection
from .types import Nodata

# pylint: disable=import-outside-toplevel


def is_rgb(x: xr.DataArray):
    """
    Check if array is RGB(A).
    """
    if x.dtype != "uint8":
        return False
    if x.ndim < 3:
        return False
    if x.shape[-1] not in (3, 4):
        return False
    return True


def _guess_rgb_names(bands: List[str]) -> Tuple[str, str, str]:
    def _candidate(color: str) -> str:
        candidates = [name for name in bands if color in name]
        n = len(candidates)
        if n == 1:
            return candidates[0]

        if n == 0:
            raise ValueError(f'Found no candidate for color "{color}"')
        raise ValueError(f'Found too many candidates for color "{color}"')

    r, g, b = [_candidate(c) for c in ("red", "green", "blue")]
    return (r, g, b)


def _auto_guess_clamp(ds: xr.Dataset) -> Tuple[float, float]:
    # TODO: deal with nodata > 0 case
    return (float(0), max(x.data.max() for x in ds.data_vars.values()))


def _to_u8(x: np.ndarray, vmin: float, vmax: float) -> np.ndarray:
    x = np.clip(x, vmin, vmax)

    if x.dtype.kind == "f":
        x = (x - vmin) * (255.0 / (vmax - vmin))
    else:
        x = (x - vmin).astype("uint32") * 255 // (vmax - vmin)
    return x.astype("uint8")


def _np_to_rgba(
    r: np.ndarray,
    g: np.ndarray,
    b: np.ndarray,
    nodata: Nodata,
    vmin: float,
    vmax: float,
) -> np.ndarray:
    rgba = np.zeros((*r.shape, 4), dtype="uint8")

    if r.dtype.kind == "f":
        valid = ~np.isnan(r)
        if nodata is not None and not np.isnan(nodata):
            valid = valid * (r != nodata)
    elif nodata is not None:
        valid = r != nodata
    else:
        valid = np.ones(r.shape, dtype=np.bool_)

    rgba[..., 3] = valid.astype("uint8") * (0xFF)
    for idx, band in enumerate([r, g, b]):
        rgba[..., idx] = _to_u8(band, vmin, vmax)

    return rgba


def to_rgba(
    ds: Any,
    bands: Optional[Tuple[str, str, str]] = None,
    *,
    vmin: Optional[float] = None,
    vmax: Optional[float] = None,
) -> xr.DataArray:
    """
    Convert dataset to RGBA image.

    Given :py:class:`xarray.Dataset` with bands ``red,green,blue`` construct :py:class:`xarray.Datarray`
    containing ``uint8`` rgba image.

    :param ds: :py:class:`xarray.Dataset`
    :param vmin: Defaults to ``0`` when ``vmax`` is supplied.
    :param vmax:
       Configure range, must be supplied for Dask inputs. When not configured
       ``vmin=0, vmax=max(r,g,b))`` is used.

    :param bands: Which bands to use, order should be red,green,blue
    """
    # pylint: disable=too-many-locals
    assert isinstance(ds, xr.Dataset)

    if bands is None:
        try:
            bands = _guess_rgb_names(list(ds.data_vars))
        except ValueError as e:
            raise ValueError(
                f"Unable to automatically guess RGB colours ({e}). "
                f"Manually specify red, green and blue bands using the "
                f"`bands` parameter."
            ) from e

    is_dask = is_dask_collection(ds)
    if vmin is None:
        if vmax is not None:
            vmin = 0

    if vmax is None:
        if is_dask:
            raise ValueError("Must specify clamp for Dask inputs (e.g. vmax, vmin)")
        _vmin, vmax = _auto_guess_clamp(ds[list(bands)])
        vmin = _vmin if vmin is None else vmin

    assert vmin is not None
    assert vmax is not None

    _b = ds[bands[0]]
    nodata = _b.odc.nodata
    dims = (*_b.dims, "band")

    r, g, b = (ds[name].data for name in bands)
    if is_dask:
        # pylint: disable=import-outside-toplevel
        from dask import array as da
        from dask.base import tokenize

        assert _b.chunks is not None
        data = da.map_blocks(
            _np_to_rgba,
            r,
            g,
            b,
            nodata,
            vmin,
            vmax,
            name=f"ro_rgba-{tokenize(r, g, b)}",
            dtype=np.uint8,
            chunks=(*_b.chunks, (4,)),
            new_axis=[r.ndim],
        )
    else:
        data = _np_to_rgba(r, g, b, nodata, vmin, vmax)

    coords = dict(_b.coords.items())
    coords.update(band=xr.DataArray(data=["r", "g", "b", "a"], dims=("band",)))

    rgba = xr.DataArray(data, coords=coords, dims=dims)
    return rgba


def _np_colorize(x, cmap, clip):
    if x.dtype == "bool":
        x = x.astype("uint8")
    if clip:
        x = np.clip(x, 0, cmap.shape[0] - 1)
    return cmap[x]


def _matplotlib_colorize(
    x,
    cmap,
    vmin=None,
    vmax=None,
    nodata: Nodata = None,
    robust=False,
):
    from matplotlib import colormaps
    from matplotlib.colors import Normalize

    if cmap is None or isinstance(cmap, str):
        # None is a valid input, maps to default cmap
        cmap = colormaps.get_cmap(cmap)  # type: ignore

    if nodata is not None:
        x = np.where(x == nodata, np.float32("nan"), x)

    if robust:
        if x.dtype.kind != "f":
            x = x.astype("float32")
        _vmin, _vmax = np.nanpercentile(x, [2, 98])

        # do not override configured values
        if vmin is None:
            vmin = _vmin
        if vmax is None:
            vmax = _vmax
    elif x.dtype.kind == "f":
        if vmin is None:
            vmin = np.nanmin(x)

        if vmax is None:
            vmax = np.nanmax(x)

    return cmap(Normalize(vmin=vmin, vmax=vmax)(x), bytes=True)


def colorize(
    x: Any,
    cmap=None,
    attrs=None,
    *,
    clip: bool = False,
    vmin: Optional[float] = None,
    vmax: Optional[float] = None,
    robust: Optional[bool] = None,
) -> xr.DataArray:
    """
    Apply colormap to data.

    There are two modes of operation:

    * Map categorical values from ``x`` to RGBA according to ``cmap`` lookup table.
    * Interpolate into RGBA using matplotlib colormaps (needs matplotlib installed)

    .. note::

       When using matplotlib colormaps with Dask inputs one must configure
       vmin/vmax to ensure all chunks are colorized consistently.


    :param x: Input xarray data array (can be Dask)
    :param cmap: Lookup table ``cmap[x] -> RGB(A)`` or matplotlib colormap
    :param vmin: Valid range to colorize
    :param vmax: Valid range to colorize
    :param robust: Use percentiles for clamping ``vmin=2%, vmax=98%``
    :param attrs: xarray attributes table, if not supplied input attributes are copied across
    :param clip: If ``True`` clip values from ``x`` to be in the safe range for ``cmap``.
    """
    # pylint: disable=too-many-locals
    from ._xr_interop import ODCExtensionDa

    assert isinstance(x, xr.DataArray)
    assert isinstance(x.odc, ODCExtensionDa)

    _is_dask = is_dask_collection(x.data)

    if isinstance(cmap, np.ndarray):
        assert cmap.ndim == 2
        assert cmap.shape[1] in (3, 4)
        cmap_dtype = cmap.dtype
        _impl = functools.partial(_np_colorize, clip=clip)
        nc = cmap.shape[1]
    else:
        # Assume matplotlib

        # default robust=True for float, non-dask inputs when vmin/vmax/robust are not configured
        if (
            vmin is None
            and vmax is None
            and robust is None
            and x.dtype.kind == "f"
            and not _is_dask
        ):
            robust = True
        elif robust is None:
            robust = False

        _impl = functools.partial(
            _matplotlib_colorize,
            vmin=vmin,
            vmax=vmax,
            nodata=x.odc.nodata,
            robust=robust,
        )
        nc, cmap_dtype = 4, "uint8"

    if attrs is None:
        attrs = {**x.attrs}
        attrs.pop("nodata", None)

    dims = (*x.dims, "band")
    coords = dict(x.coords.items())
    coords["band"] = xr.DataArray(data=["r", "g", "b", "a"][:nc], dims=("band",))

    if _is_dask:
        from dask import array as da
        from dask import delayed
        from dask.base import tokenize

        _cmap = delayed(cmap) if isinstance(cmap, np.ndarray) else cmap

        assert x.chunks is not None
        data = da.map_blocks(
            _impl,
            x.data,
            _cmap,
            name=f"colorize-{tokenize(x, _cmap, clip, vmin, vmax, robust)}",
            meta=np.ndarray((), cmap_dtype),
            chunks=(*x.chunks, (nc,)),
            new_axis=[x.data.ndim],
        )
    else:
        data = _impl(x.data, cmap)

    return xr.DataArray(data=data, dims=dims, coords=coords, attrs=attrs)


def replace_transparent_pixels(
    rgba: np.ndarray, color: Tuple[int, int, int] = (255, 0, 255)
) -> np.ndarray:
    """
    Convert RGBA to RGB.

    Replaces transparent pixels with a given color.
    """
    assert rgba.ndim == 3
    assert rgba.shape[-1] == 4

    m = rgba[..., -1] == 0
    rgb = rgba[..., :3].copy()
    rgb[m] = color
    return rgb
